# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Evan Wojciechowski
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import matplotlib.mlab as mlab
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from .vis_utils import *
    # mod_data() for a (more) thorough checking routine than in other functions, keeping plot_spiral from breaking

def plot_spiral(plotData, plotIndex, T, mainTitle = 'Spiral plot', barLabel = 'Amplitude', plotTS = False, show = True):
    '''
    Plots data in a spiral pattern via a polar plot

    @param plotData: Input data values/amplitudes
    @param plotIndex: Input index (series time coordinates)
    @param period: Period value with which to wrap data around the plot
    '''

    r     = plotIndex - plotIndex.min()
    theta = 2 * np.pi * r / T
    z     = plotData

    fig = plt.figure()
    ax  = fig.add_axes([.1, .1, .8, .8], projection = 'polar', frameon = False)
    ax.set_title(mainTitle, fontsize = 18)

    heatMap = ax.scatter(theta, r, c = z, s = 10, edgecolors = 'none', cmap = 'plasma')
    cbar = fig.colorbar(heatMap)
    cbar.set_label(barLabel, fontsize = 18)
    cbar.ax.tick_params(labelsize = 12)

    ax.set_rmax(max(r))
    ax.set_rlabel_position(-22.5)
    ax.set_xticks(np.pi/180. * np.linspace(0, 360, 8, endpoint = False))
    ax.grid(True)

    ax.text(0, 0, 'Period = {:.4f}'.format(T), fontsize = 14, transform = ax.transAxes)

    if plotTS: # only works for kepler, basically broken for others
        plt.figure()
        plt.plot(plotIndex % T, plotData, '.')
    
    if show:
        plt.show()

#%# wrappers #%#
def run_spiral(inData, period, inIndex = None, mainTitle = 'Spiral plot', barLabel = 'Amplitude', plotTS = False, show = True):
    '''
    Wrapper for plot_spiral

    @param inData: Input data to use in plot
    @param period: Period value with which to wrap data around the plot
    @param inIndex: Input index (series time coordinates)
    @param plotTS: Optional flag to plot the time series of the data in a separate window
    @param show: Boolean to show plot immediately after plot creation
    '''

    plotData, plotIndex = mod_data(inData, inIndex, makeType = int)
    plot_spiral(plotData, plotIndex, T = period, mainTitle = mainTitle, barLabel = barLabel, plotTS = plotTS, show = show)

def run_spiralInteractive(inData, period, pParams = [], inIndex = None, mainTitle = 'Spiral plot', barLabel = 'Amplitude', plotTS = False):
    '''
    Wrapper for plot_spiral that is interactive when used in Jupyter notebooks

    @param inData: Input data to use in plot
    @param period: Period value with which to wrap data around the plot
    @param pParams: List of plot's period parameters [min, max, step] necessary for interactive
    @param inIndex: Input index (series time coordinates)
    @param plotTS: Optional flag to plot the time series of the data in a separate window
    '''

    from ipywidgets import fixed, interact, interactive
    from IPython.display import clear_output, display, HTML
    
    plotData, plotIndex = mod_data(inData, inIndex)

    if len(pParams) != 3:
        pParams = [period - .1, period + .1, .01]

    inter = interactive(plot_spiral, plotData = fixed(plotData), plotIndex = fixed(plotIndex), mainTitle = fixed(mainTitle), barLabel = fixed(barLabel), plotTS = fixed(plotTS), T = (pParams[0], pParams[1], pParams[2]))
    display(inter)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from six.moves import configparser
import os

def getConfig():
    '''
    Retrieve skdiscovery configuaration

    @return skdiscovery configparser
    '''

    config_location = os.path.expanduser('~') + '/.skdiscovery.conf'

    conf = configparser.ConfigParser()
    conf.read(config_location)

    return conf

def writeConfigValue(section,key,value):
    '''
    Write config to disk

    @param section: Name of section
    @param key: Name of key
    @param value: Value to write
    '''

    conf = getConfig()

    if not conf.has_section(section):
        conf.add_section(section)

    conf.set(section, key, value)    
    
    config_location = os.path.expanduser('~') + '/.skdiscovery.conf'
    config_handle = open(config_location, "w")
    conf.write(config_handle)
    config_handle.close()


def getDispyPassword():
    '''
    Get dispy password

    @return dispy password
    '''
    conf = getConfig()

    return conf.get('Dispy', 'password', fallback=None)

def getHostName():
    '''
    Get Host name for displaying link to dispy status

    @return Hostname
    '''
    conf = getConfig()

    return conf.get('Dispy','hostname')



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem
import pandas as pd
import numpy as np


class Correlate(PipelineItem):
    '''
    Computes the correlation for series data

    Stores the result as a matrix
    '''

    def __init__(self, str_description, labels=None, column_names=None):
        '''
        Initialize Correlate analysis item

        @param str_description: String describing analysis item
        @param labels: List of labels used to select data
        @param column_names: List of column names used to select data
        '''
        
        super(Correlate, self).__init__(str_description,[])
        self.labels = labels
        self.column_names = column_names
    
    def process(self, obj_data):
        ''' 
        Computes the correlation between all the time series

        The results are stored in obj_data

        @param obj_data: Data wrapper for correlating
        '''
        data = []
        index = []

        for label, series, err in obj_data.getIterator():
            if (self.labels is None or label in self.labels) and \
               (self.column_names is None or series.name in self.column_names):

                data.append(series)
                index.append(series.name)

        index = np.array(index)
        

        result = []
        for s1 in data:
            row = []
            for s2 in data:
                row.append(s1.corr(s2))
            result.append(row)

        obj_data.addResult(self.str_description, pd.DataFrame(result, index=index, columns=index))



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software has been created in projects supported by the US National
# Science Foundation and NASA (PI: Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework.base import PipelineItem
import numpy as np
import pandas as pd
from statsmodels.robust import mad


class MIDAS(PipelineItem):
    '''
    *In Development* A basic MIDAS trend estimator

    See http://onlinelibrary.wiley.com/doi/10.1002/2015JB012552/full
    '''
        
    def __init__(self, str_description,column_names = None):
        '''
        Initiatlize the MIDAS filtering item

        @param str_description: String description of filter
        @param column_names: List of column names to analyze
        '''
        
        super(MIDAS, self).__init__(str_description, [])
        self.column_names = column_names

    def process(self, obj_data):
        '''
        Apply the MIDAS estimator to generate velocity estimates

        Adds the result to the data wrapper

        @param obj_data: Data wrapper
        '''
        
        if self.column_names == None:
            column_names = obj_data.getDefaultColumns()
        else:
            column_names = self.column_names

        time_diff = pd.to_timedelta('365d')
        results = dict()
        for label, data in obj_data.getIterator():
            start_date = data.index[0]
            end_date = data.index[-1]
            for column in column_names:
                start_data = data.loc[start_date:(end_date-time_diff), column]
                end_data = data.loc[start_date+time_diff:end_date, column]
                
                offsets = end_data.values - start_data.values
                offsets = offsets[~np.isnan(offsets)]
                med_off = np.median(offsets)
                mad_off = mad(offsets)
                
                cut_offsets = offsets[np.logical_and(offsets < med_off + 2*mad_off, 
                                                     offsets > med_off - 2*mad_off)]
                final_vel = np.median(cut_offsets)
                final_unc = np.sqrt(np.pi/2) * mad(cut_offsets) / np.sqrt(len(cut_offsets))

                results[label] = pd.DataFrame([final_vel,final_unc], ['velocity', 'uncertainty'] ,[column])
                
        obj_data.addResult(self.str_description, pd.Panel.fromDict(results,orient='minor'))
        




# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Evan Wojciechowski
# This software is part of projects sponsored by NASA and NSF (PI: V. Pankratius)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# linear interpolation functions and plotting
# mostly for use in tandem with other functions

import matplotlib.pyplot as plt
import numpy as np

from .vis_utils import coldict


def lin_trend(inData):
    '''
    Calculates a linear polynomial fit and evaluates

    @param inData: Input data to fit

    @return Array of evaluated points for the linear fit
    '''

    real = np.isfinite(inData)
    rng  = np.arange(len(inData))

    lin_fit = np.poly1d(np.polyfit(rng[real], inData[real], 1))(rng[real])

    return lin_fit

def calc_lin_interp(inData, iterStep = 100):
    '''
    Calculates a piecewise linear interpolated fit for some data

    @param inData: Input data to fit
    @param iterStep: Number of data points per interpolation step

    @return Array of interpolated values
    '''

    if iterStep > len(inData):
        iterStep = input('Step size too large (need < {}), enter new: '.format(len(inData)))

    intIndex = np.arange(len(inData))
    interp = np.interp(intIndex[::iterStep], intIndex, inData) 
    return interp

def plot_lin_trend(inData, plotIndex = None, show = True):
    '''
    Plots a linear linear trend against its source data

    @param inData: Input data to fit and plot
    @param plotIndex: Optional index array to pass for plotting
    @param show: Boolean to show plot immediately after plot creation
    '''

    real = np.isfinite(inData)

    if plotIndex is None:
        try:
            plotIndex = inData[real].index
        except:
            plotIndex = np.arange(len(inData[real]))

    lin_fit = lin_trend(inData)

    plt.plot(plotIndex, inData, color = coldict[0])
    plt.plot(plotIndex, lin_fit, color = coldict[0], linestyle = '--', alpha = .4)

    if show:
        plt.show()

def plot_lin_interp(inData, interps = None, plotIndex = None, iterSteps = [100], pRange = [], mainTitle = 'Piecewise Decomposition', plotReal = True, show = True):
    '''
    Plots linear interpolation against its source data

    @param inData: Input data to fit and plot
    @param interps: Optional interpolated data to be plotted, will be made if not given
    @param plotIndex: Optional index array to pass for plotting
    @param iterSteps: List of iterStep values to calculate/plot
    @param pRange: Range over which to plot, defaults to start and end of original data
    @param mainTitle: Optional string plot title
    @param plotReal: Boolean variable to optionally disinclude source data
    @param show: Boolean to show plot immediately after plot creation

    @return Multidimensional array of interpreted data values
    '''

    col = 0

    fig = plt.figure(figsize = (12, 6.75))

    if interps is None:
        interps = [calc_lin_interp(inData, step) for step in iterSteps]

    if len(pRange) != 2:
        pRange = [0, len(inData)]

    if plotIndex is None:
        try:
            plotIndex = inData.index
            if ((plotIndex.dtype == np.dtype('<M8[ns]')) or (plotIndex.dtype == np.dtype('datetime64[ns]'))):
                xlabel = 'Time'
            else:
                xlabel = 'Index'
        except:
            plotIndex = np.arange(len(inData))
            xlabel    = 'Index'
    else:
        if ((plotIndex.dtype == np.dtype('<M8[ns]')) or (plotIndex.dtype == np.dtype('datetime64[ns]'))):
             xlabel = 'Time'
        else:
             xlabel = 'Index'

    if plotReal:
        plt.plot(plotIndex[pRange[0] : pRange[1]], inData[pRange[0] : pRange[1]], color = '#CC0000', label = 'Raw data')

    for num, step in enumerate(iterSteps):
        plt.plot(plotIndex[pRange[0] : pRange[1] : step], interps[num][pRange[0] : pRange[1]], color = coldict[col % 16], label = 'Interpolated ({} points/segment)'.format(step))
        col += 1

    plt.legend(fontsize = 14)
    plt.xlabel(xlabel, fontsize = 18)
    plt.ylabel('Amplitude', fontsize = 18)
    plt.title(mainTitle, fontsize = 18)
    plt.xticks(fontsize = 18)
    plt.yticks(fontsize = 18)

    if show:
        plt.show()

    return interps


def plot_lin_slope(inData, interps = None, plotIndex = None, mainTitle = 'Piecewise Decomposition and Slopes', iterSteps = [100], pRange = [], plotReal = True, show = True):
    '''
    Plots raw data, linear interpolated data, and interpolated slope

    @param inData: Input data to fit and plot
    @param interps: Optional interpolated data to be plotted, will be made if not given
    @param plotIndex: Optional index array to pass for plotting
    @param iterSteps: List of iterStep values to calculate/plot
    @param pRange: Range over which to plot, defaults to start and end of original data
    @param mainTitle: Optional string plot title
    @param plotReal: Boolean variable to optionally disinclude source data
    @param show: Boolean to show plot immediately after plot creation   

    @return Tuple of interpolated values array and corresponding gradient array
    '''

    coli = 0; colg = 0

    if interps is None:
        interps = [calc_lin_interp(inData, step) for step in iterSteps]

    if len(pRange) != 2:
        pRange = [0, len(inData)]

    if plotIndex is None:
        try:
            plotIndex = inData.index
            if ((plotIndex.dtype == np.dtype('<M8[ns]')) or (plotIndex.dtype == np.dtype('datetime64[ns]'))):
                xlabel = 'Time'
            else:
                xlabel = 'Index'
        except:
            plotIndex = np.arange(len(inData))
            xlabel    = 'Index'
    else:
        if ((plotIndex.dtype == np.dtype('<M8[ns]')) or (plotIndex.dtype == np.dtype('datetime64[ns]'))):
             xlabel = 'Time'
        else:
             xlabel = 'Index'

    gradients = [np.gradient(interp) for interp in interps]

    fig, axes = plt.subplots(2 + plotReal, 1, sharex = True, figsize = (12, 6.75))
    
    fig.add_subplot(111, frameon = False)
    plt.tick_params(labelcolor = 'none', top = 'off', bottom = 'off', left = 'off', right = 'off')
    plt.xlabel(xlabel, fontsize = 18)
    plt.ylabel('Amplitude', fontsize = 18)

    if plotReal:
        axes[0].plot(plotIndex, inData, color = '#CC0000')
        axes[0].set_title(mainTitle, fontsize = 18)

    for num, interp in enumerate(interps):
        axes[0 + plotReal].plot(plotIndex[::iterSteps[num]], interp, color = coldict[coli % 16], label = '{} points/segment'.format(iterSteps[num]))
        axes[0 + plotReal].set_title('Interpolated', fontsize = 18)
        coli += 1
    axes[0 + plotReal].legend(fontsize = 12)

    for num, grad in enumerate(gradients):
        axes[1 + plotReal].plot(plotIndex[::iterSteps[num]], grad, color = coldict[colg % 16], label = '{} points/segment'.format(iterSteps[num]))
        axes[1 + plotReal].set_title('Interpolated slope', fontsize = 18)
        colg += 1
    axes[1 + plotReal].legend(fontsize = 12)

    for ax in axes:
        ax.tick_params(axis = 'both', labelsize = 14)
    
    if show:
        plt.show()

    return interps, gradients

"""
def calc_gradient_extrema(gradient, numExt = 3):
    '''
    NOT DONE
    '''

    #! make it calculate where on original index extrema lie
    
    arr = gradient.copy()
    maxs = {}
    mins = {}
    
    for _ in np.arange(numExt):
        maxs[np.argmax(arr)] = np.max(arr)
        mins[np.argmin(arr)] = np.min(arr)
        arr = np.delete(arr, (np.argmax(arr), np.argmin(arr)))

    extrema = np.concatenate((list(maxs.values()), list(mins.values())))
    indices = np.concatenate((list(maxs.keys()), list(mins.keys())))

    return indices, extrema
"""



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Library imports
import pandas as pd
import numpy as np

# skdiscovery imports
from skdiscovery.data_structure.framework.base import PipelineItem


class PropagateNaNs(PipelineItem):
    ''' Propagates NaN's from one column to other columns '''
    
    def __init__(self,str_description,nan_column,target_columns):
        '''
        Initialize PropagateNaNs Filter

        @param str_description: String describing filter
        @param nan_column: Column used to select which rows should be NaN's
        @param target_columns: Rows in these column will be set to NaN's based on nan_column
        '''

        super(PropagateNaNs,self).__init__(str_description,[])
        self.nan_column = nan_column
        self.target_columns = target_columns
        
    def process(self, obj_data):
        '''
        PropagateNaNs on table data wrapper

        @param obj_data: Input table data wrapper
        '''
        
        for label, data in obj_data.getIterator():
            nan_index = data[pd.isnull(data[self.nan_column])].index
            
            for column in self.target_columns:
                new_data = data[column].copy()
                new_data[nan_index] = np.nan
                
                obj_data.updateData(label,new_data.index,column,new_data)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem
import scipy as sp
import matplotlib.pyplot as plt


class HCluster(PipelineItem):
    '''
    Hierarchical Clustering function that produces a cluster map of the distance matrix
    '''
    def __init__(self, str_description, obj_name):
        '''
        Initialize HCluster

        @param str_description: String describing accumulator
        @param obj_name: Name of distance matrix parameter in the obj_data results
        '''
        
        super(HCluster, self).__init__(str_description)
        self.obj_name = obj_name
        
    def process(self, obj_data):
        '''
        Produces a cluster map and stores the linkage results.
    
        @param obj_data: Data wrapper
        '''

        import seaborn as sns
        
        data = obj_data.getResults()[self.obj_name]

        linkage = sp.cluster.hierarchy.linkage(data, method='average')

        plt.figure()
        
        g = sns.clustermap(data, col_linkage = linkage, row_linkage=linkage)
        for item in g.ax_heatmap.get_yticklabels():
            item.set_rotation(0)
        

        plt.figure()

        sp.cluster.hierarchy.dendrogram(linkage)

        obj_data.addResult(self.str_description, linkage)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Author: Cody
# This software has been created in projects supported by the US National
# Science Foundation and NASA (PI: Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from skdiscovery.data_structure.framework import PipelineItem
from sklearn.cluster import DBSCAN

class DBScan(PipelineItem):
    '''
    Runs DBScan on table data

    Adds cluster information column to data
    '''

    def __init__(self, str_description, ap_paramList, column_names):
        '''
        Initialize DBScan pipelne item

        @param str_description: Description of item
        @param ap_paramList[epsilon]: Distance between two nodes for them to be considered connected
        @param ap_paramList[min_points]: Minimum number of points for a cluster
        @param column_names: List of column names to use
        '''
        self.column_names = column_names
        super(DBScan, self).__init__(str_description, ap_paramList)
    
    def process(self, obj_data):
        ''' 
        Run DBScan on data. Stores result in data wrapper

        @param obj_data: Data wrapper to be processed
        '''

        epsilon = self.ap_paramList[0]()
        min_points = self.ap_paramList[1]()

        results = dict()
        

        for label, data in obj_data.getIterator():
            results[label] = DBSCAN(eps=epsilon, min_samples = min_points).fit_predict(data.loc[:,self.column_names])

        obj_data.addResult(self.str_description, results)



from .correlate import Correlate
from .gca import General_Component_Analysis
from .mogi import Mogi_Inversion
from .skew import Skew
from .midas import MIDAS
from .outlier import Outlier
# from .max_entropy_detect import MaxEntropyDetector
# from .entropy_correlation_classify import EntropyCorrelationClassifier
# from .neural_network_classify import NeuralNetworkClassifier
# from .random_forest_classify import RandomForestClassifier



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem
import pandas as pd
import numpy as np


class Correlate(PipelineItem):
    '''
    Computes the correlation for table data and stores the result as a matrix.
    '''

    def __init__(self, str_description, column_names = None, local_match = False, correlation_type = 'pearson'):
        '''
        Initialize Correlate analysis item for use on tables

        @param str_description: String describing analysis item
        @param column_names: List of column names to correlate
        @param local_match: Only correlate data on the same frames
        @param correlation_type: Type of correlation to be passed to pandas ('pearson', 'kendall', 'spearman')
        '''
        super(Correlate, self).__init__(str_description,[])
        self.column_names = column_names
        self.local_match = local_match
        self.corr_type = correlation_type
    
    
    def process(self, obj_data):
        ''' 
        Computes the correlation between columns and stores the results in obj_data

        @param obj_data: Data wrapper
        '''

        if self.column_names == None:
            column_names = obj_data.getDefaultColumns()
        else:
            column_names = self.column_names
        
        if self.local_match == False:
            data = []
            index = []

            for label, data_in in obj_data.getIterator():
                for column in column_names:
                    data.append(data_in[column])
                    index.append(label + '.' + column)

            index = np.array(index)


            result = []
            for s1 in data:
                row = []
                for s2 in data:
                    row.append(s1.corr(s2, method=self.corr_type))
                result.append(row)

            obj_data.addResult(self.str_description, pd.DataFrame(result, index=index, columns=index))

        else:
            full_results = dict()
            for label, data_in in obj_data.getIterator():
                data = []
                index = []
                for column in column_names:
                    data.append(data_in[column])
                    index.append(column)

                result = []
                for s1 in data:
                    row = []
                    for s2 in data:
                        row.append(s1.corr(s2, method=self.corr_type))
                    result.append(row)

                full_results[label] = pd.DataFrame(result,index=index,columns=index)

            obj_data.addResult(self.str_description, pd.Panel.from_dict(full_results))



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import skdaccess.utilities.pbo_util as pbo_tools
import skdiscovery.data_structure.series.analysis.mogi as mogi

from mpl_toolkits.basemap import Basemap
import numpy as np
import matplotlib.pyplot as plt



def multiCaPlot(pipeline, mogiFlag=False, offset=.15, direction='H',pca_comp=0,scaleFactor=2.5,map_res='i'):
    '''
    The multiCaPlot function generates a geographic eigenvector plot of several pipeline runs
    
    This function plots multiple pipeline runs over perturbed pipeline
    parameters. The various perturbations are plotted more
    transparently (alpha=.5), while the median eigen_vector and Mogi
    inversion are plotted in solid blue and red

    @param pipeline: The pipeline object with multiple runs
    @param mogiFlag: Flag to indicate plotting the Mogi source as well as the PCA
    @param offset: Offset for padding the corners of the generated map
    @param direction: Indicates the eigenvectors to plot. Only Horizontal component is currently supported ('H')
    @param pca_comp: Choose the PCA component to use (integer)
    @param scaleFactor: Size of the arrow scaling factor
    @map_res: Map data resolution for Basemap ('c', 'i', 'h', 'f', or None)
    '''
    
    # as this is a multi_ca_plot function, assumes GPCA
    plt.figure();
    
    meta_data = pipeline.data_generator.meta_data
    station_list = pipeline.data_generator.station_list
    
    lat_range, lon_range = pbo_tools.getLatLonRange(meta_data, station_list)
    coord_list = pbo_tools.getStationCoords(meta_data, station_list)
    
    # Create a map projection of area
    bmap = Basemap(llcrnrlat=lat_range[0] - offset, urcrnrlat=lat_range[1] + offset, llcrnrlon=lon_range[0] - offset, urcrnrlon=lon_range[1] + offset,
               projection='gnom', lon_0=np.mean(lon_range), lat_0=np.mean(lat_range), resolution=map_res)
    
    # bmap.fillcontinents(color='white')
    # bmap.drawmapboundary(fill_color='white')
    bmap.drawmapboundary(fill_color='#41BBEC');
    bmap.fillcontinents(color='white')
    
    # Draw just coastlines, no lakes
    for i,cp in enumerate(bmap.coastpolygons):
        if bmap.coastpolygontypes[i]<2:
            bmap.plot(cp[0],cp[1],'k-')
    
    parallels = np.arange(np.round(lat_range[0]-offset,decimals=1),np.round(lat_range[1]+offset,decimals=1),.1)
    meridians = np.arange(np.round(lon_range[0]-offset,decimals=1),np.round(lon_range[1]+offset,decimals=1),.1)
    
    bmap.drawmeridians(meridians, labels=[0,0,0,1])
    bmap.drawparallels(parallels, labels=[1,0,0,0])
    
    # Plot station coords
    for coord in coord_list:
        bmap.plot(coord[1], coord[0], 'ko', markersize=6, latlon=True,zorder=12)
        x,y = bmap(coord[1], coord[0])
        plt.text(x+250,y-450,station_list[coord_list.index(coord)],zorder=12)
    
    # loop over each pipeline run
    elatmean = np.zeros(len(station_list))
    elonmean = np.zeros_like(elatmean)
    # check if want to plot Mogi as well
    if mogiFlag:
        avg_mogi = np.array([0.,0.])
        mlatmean = np.zeros_like(elatmean)
        mlonmean = np.zeros_like(elatmean)
    
    for nrun in range(len(pipeline.RA_results)):
        pca = pipeline.RA_results[nrun]['GPCA']['CA']
        station_lat_list, station_lon_list, ev_lat_list, ev_lon_list, dir_sign = pbo_tools.dirEigenvectors(coord_list, pca.components_[pca_comp])
    
        elatmean += ev_lat_list
        elonmean += ev_lon_list
        # plot each run in light blue
        bmap.quiver(station_lon_list, station_lat_list, ev_lon_list, ev_lat_list, latlon=True,
                    scale = scaleFactor, alpha = .25, color = 'blue',zorder=11)
    
        if mogiFlag:
            mogi_res = pipeline.RA_results[nrun]['Mogi']
            avg_mogi += np.array([mogi_res['lon'], mogi_res['lat']])
            mogi_x_disp, mogi_y_disp = mogi.MogiVectors(mogi_res,station_lat_list,station_lon_list)
            mlatmean += mogi_y_disp
            mlonmean += mogi_x_disp
    
            bmap.plot(mogi_res['lon'], mogi_res['lat'], "g^", markersize = 10, latlon=True, alpha = .25,zorder=12)
            bmap.quiver(station_lon_list, station_lat_list, mogi_x_disp*dir_sign, mogi_y_disp*dir_sign,
                        latlon=True, scale=scaleFactor,color='red', alpha = .25,zorder=11)
    
    #plot the mean ev in blue
    elatmean = elatmean/len(pipeline.RA_results)
    elonmean = elonmean/len(pipeline.RA_results)
    bmap.quiver(station_lon_list, station_lat_list, elonmean, elatmean,
                latlon=True, scale = scaleFactor, color = 'blue', alpha = 1,zorder=11)
    if mogiFlag:
        # plot mean mogi results
        avg_mogi = avg_mogi/len(pipeline.RA_results)
        mlatmean = mlatmean/len(pipeline.RA_results)
        mlonmean = mlonmean/len(pipeline.RA_results)
        bmap.plot(avg_mogi[0], avg_mogi[1], "g^", markersize = 10, latlon=True, alpha = 1,zorder=12)
        bmap.quiver(station_lon_list, station_lat_list, mlonmean*dir_sign, mlatmean*dir_sign,
                    latlon=True, scale=scaleFactor,color='red', alpha = 1,zorder=11)
    
    
    ax_x = plt.gca().get_xlim()
    ax_y = plt.gca().get_ylim()
    x,y = bmap(ax_x[0]+.1*(ax_x[1]-ax_x[0]), ax_y[0]+.1*(ax_y[1]-ax_y[0]),inverse=True)
    bmap.quiver(x, y, 0, .2, latlon=True, scale = scaleFactor, headwidth=3,headlength=3,zorder=11)
    plt.text(ax_x[0]+.1*(ax_x[1]-ax_x[0])-650, ax_y[0]+.1*(ax_y[1]-ax_y[0])-1000,'20%',zorder=11)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Evan Wojciechowski
# This software is part of projects sponsored by NASA and NSF (PI: V. Pankratius)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# emd (empirical mode decomposition) and hht (hilbert huang transform) functions
#! robust up until ~1500 data points
#! rename this file?
#! need to look over wrappers, streamline

import os, sys, warnings

from matplotlib.ticker import FormatStrFormatter
from scipy import stats
import matplotlib.mlab as mlab
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyhht

from .vis_utils import *
    # block_output(), enable_output(),


# core functionality

def calc_imfs(rawData, nbsym = False):
    '''
    IMF calculation function, streamlined and quieted

    @param rawData: Input data for EMD calculation
    @param nbsym: Boolean that would add extra data points near boundaries when calculating; breaks some datasets unless False

    @return 2D numpy.ndarray of IMFs
    '''

    block_output()
    imfs = pyhht.emd.EMD(rawData, nbsym = nbsym).decompose()
    enable_output()
    return imfs

def calc_imfs_sum(imfs, highNum = 2, high = True, residual = False):
    '''
    IMF summation helper function

    @param imfs: Input array of IMFs to be summed
    @param highNum: Number of high frequency IMFs to sum, starting from IMF1 (indexed at 0)
    @param high: Boolean that determines which class of frequency to sum (default True to sum HF)
    @param residual: Boolean that optionally includes the residual function when summing low frequency IMFs (default False to disinclude residual)

    @return 1D numpy.ndarray of summed IMFs
    '''

    if high:
        summed = np.sum(imfs[: highNum], axis = 0)
    else:
        summed = np.sum(imfs[highNum : len(imfs) - residual], axis = 0)

    return summed

def plot_imfs(rawData, imfs, toPlot = [], mainTitle = 'IMFs', show = True, figsize=(12,10)):
    '''
    Plots raw data and IMFs in a subplot grid (n Imfs [rows] x 1 [col])
    
    @param rawData: Input data for plotting
    @param imfs: Input array of IMFs for plotting
    @param toPlot: List of which IMFs to plot (default is all)
    @param show: Boolean to show plot immediately after plot creation
    @param figsize: Size of figure
    '''

    try:
        plotIndex = rawData.index
        if ((plotIndex.dtype == np.dtype('<M8[ns]')) or (plotIndex.dtype == np.dtype('datetime64[ns]'))):
            xlabel = 'Time'
        else:
            xlabel = 'Index'

    except:
        plotIndex = np.arange(len(rawData))
        xlabel    = 'Index'

    if len(toPlot) == 0:
        toPlot = np.arange(len(imfs))
    else:
        toPlot = [i - 1 for i in toPlot]

    fig, axes = plt.subplots(len(toPlot) + 1, 1, sharex = True, figsize = figsize)

    naxis = fig.add_subplot(111, frameon = False)
    plt.tick_params(labelcolor = 'none', top = 'off', bottom = 'off', left = 'off', right = 'off')

    if len(imfs) <= 10:
        lSize = 14
    else:
        lSize = 12


    axes[0].plot(plotIndex, rawData, color = '#0066FF')
    axes[0].set_title(mainTitle, fontsize = 18)
    axes[0].set_ylabel('Raw', fontsize = 18)

    axes[-1].set_xlabel(xlabel, fontsize = 18)

    axnum = 1

    for i in toPlot:
        if i != (len(imfs) - 1):
            axes[axnum].plot(plotIndex, imfs[i], color = '#009900')
            axes[axnum].set_ylabel('IMF{}'.format(i + 1), fontsize = lSize)
            axnum += 1

        else:
            axes[axnum].plot(plotIndex, imfs[-1], color = '#000000')
            axes[axnum].set_ylabel('Resid', fontsize = lSize)
            axnum += 1

    for ax in axes:
        ax.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))

        ax.tick_params(axis = 'x', labelsize = 18)
        ax.tick_params(axis = 'y', labelsize = lSize)

    if show:
        plt.show()

def plot_imfs_split(rawData, imfs, highNum = 2, residual = False, mainTitle = 'Raw data', collage = False, show = True):
    '''
    Plots raw data and summed IMFs based on HF/LF cut, can optionally plot the residual separately from LF

    @param rawData: Input data for plotting
    @param imfs: Input array of IMFs for summing and then plotting
    @param highNum: Number of high frequency IMFs to sum, starting from IMF1 (indexed at 0)
    @param residual: Boolean that optionally includes the residual function when summing low frequency IMFs (default False to disinclude residual)
    @param mainTitle: Title string of plot
    @param collage: Boolean that can optionally return certain plot parameters for external usage
    @param show: Boolean to show plot immediately after plot creation

    @return Tuple of HF summed data array and LF summed data array
    '''
    try:
        plotIndex = rawData.index
        if ((plotIndex.dtype == np.dtype('<M8[ns]')) or (plotIndex.dtype == np.dtype('datetime64[ns]'))):
            xlabel = 'Time'
        else:
            xlabel = 'Index'

    except:
        plotIndex = np.arange(len(rawData))
        xlabel    = 'Index'
    
    highData = calc_imfs_sum(imfs, highNum = highNum, residual = residual)
    lowData  = calc_imfs_sum(imfs, highNum = highNum, residual = residual, high = False)

    plotArgs = rawData, highData, lowData, imfs[-1],
    if collage:
        return plotArgs

    fig, axes = plt.subplots(3 + residual, 1, sharex = True, figsize = (12, 6.75))

    fig.add_subplot(111, frameon = False)
    plt.tick_params(labelcolor = 'none', top = 'off', bottom = 'off', left = 'off', right = 'off')
    plt.xlabel(xlabel, fontsize = 18)
    plt.ylabel('Amplitude', fontsize = 18)

    axes[0].plot(plotIndex, plotArgs[0], color = '#0066FF')
    axes[1].plot(plotIndex, plotArgs[1], color = '#CC33FF')
    axes[2].plot(plotIndex, plotArgs[2], color = '#CC0000')

    axes[0].set_title(mainTitle, fontsize = 18)
    axes[1].set_title('{} highest frequency IMFs'.format(highNum), fontsize = 18)
    if highNum == 1:
        axes[1].set_title('Highest frequency IMF', fontsize = 18)
    axes[2].set_title('{} lowest frequency IMFs and residual'.format(len(imfs) - highNum -1), fontsize = 18)
    if (len(imfs) - highNum) == 2:
            axes[2].set_title('Lowest frequency IMF and residual', fontsize = 18)

    if residual:
        axes[3].plot(plotIndex, plotArgs[3], '#000000')
            
        axes[2].set_title('{} lowest frequency IMFs'.format(len(imfs) - highNum - 1), fontsize = 18)
        if (len(imfs) - highNum) == 2:
            axes[2].set_title('Lowest frequency IMF', fontsize = 18)
        axes[3].set_title('Residual', fontsize = 18)
    
    for ax in axes:
        ax.tick_params(axis = 'x', labelsize = 12)
        ax.tick_params(axis = 'y', labelsize = 12)

    if show:
        plt.show()

    return highData, lowData


def plot_imfs_split_comp(rawData, imfs, highNums = [2, 3], residual = False, plotRaw = True, mainTitle = 'Raw data', collage = False, show = True):
    '''
    Like plot_imfs_split, plots raw data and summed IMFs based on two HF/LF cuts

    @param rawData: Input data for plotting
    @param imfs: Input array of IMFs for summing and then plotting
    @param highNums: Number of high frequency IMFs to sum and compare
    @param residual: Boolean that optionally includes the residual function when summing low frequency IMFs (default False to disinclude residual)
    @param plotRaw: Boolean to optionally disinclude raw data plot above IMF summation comparison
    @param mainTitle: Title string of plot
    @param collage: Boolean that can optionally return certain plot parameters for external usage
    @param show: Boolean to show plot immediately after plot creation

    @return Tuple of both HF summed data arrays and borh LF summed data arrays
    '''

    try:
        plotIndex = rawData.index
        if ((plotIndex.dtype == np.dtype('<M8[ns]')) or (plotIndex.dtype == np.dtype('datetime64[ns]'))):
            xlabel = 'Time'
        else:
            xlabel = 'Index'

    except:
        plotIndex = np.arange(len(rawData))
        xlabel    = 'Index'

    highData1 = calc_imfs_sum(imfs, highNum = highNums[0], residual = residual)
    highData2 = calc_imfs_sum(imfs, highNum = highNums[1], residual = residual)
    lowData1  = calc_imfs_sum(imfs, highNum = highNums[0], residual = residual, high = False)
    lowData2  = calc_imfs_sum(imfs, highNum = highNums[1], residual = residual, high = False)

    plotArgs = plotIndex, rawData, highData1, highData2, lowData1, lowData2, imfs[-1]
    if collage:
        return plotArgs

    fig = plt.figure(figsize = (12, 6.75))
    
    fig.add_subplot(111, frameon = False)
    plt.tick_params(labelcolor = 'none', top = 'off', bottom = 'off', left = 'off', right = 'off')
    plt.xlabel(xlabel, fontsize = 18)
    plt.ylabel('Amplitude', fontsize = 18)


    if plotRaw:
        ax_main = fig.add_subplot(3 + residual, 1 , 1)
        ax_main.set_title(mainTitle, fontsize = 18)
        ax_main.plot(plotArgs[0], plotArgs[1], color = '#0066FF')

    axhigh1 = fig.add_subplot(2 + plotRaw + residual, 2, 1 + 2*plotRaw)
    axhigh2 = fig.add_subplot(2 + plotRaw + residual, 2, 2 + 2*plotRaw, sharey = axhigh1)
    axlow1  = fig.add_subplot(2 + plotRaw + residual, 2, 3 + 2*plotRaw, sharex = axhigh1)
    axlow2  = fig.add_subplot(2 + plotRaw + residual, 2, 4 + 2*plotRaw, sharex = axhigh2, sharey = axlow1)

    axhigh1.set_title('{} highest frequency IMFs'.format(highNums[0]), fontsize = 16)
    axhigh2.set_title('{} highest frequency IMFs'.format(highNums[1]), fontsize = 16)
    axlow1.set_title('{} lowest frequency IMFs and residual'.format(len(imfs) - highNums[0] - 1), fontsize = 16)
    axlow2.set_title('{} lowest frequency IMFs and residual'.format(len(imfs) - highNums[1] - 1), fontsize = 16)

    axhigh1.plot(plotArgs[0], plotArgs[2], color = '#CC33FF')
    axhigh2.plot(plotArgs[0], plotArgs[3], color = '#CC33FF')
    axlow1.plot(plotArgs[0], plotArgs[4], color = '#CC0000')
    axlow2.plot(plotArgs[0], plotArgs[5], color = '#CC0000')

    axL = [ax_main, axhigh1, axhigh2, axlow1, axlow2]

    if residual:
        axresid = fig.add_subplot(2 + plotRaw + residual, 2, (7,8))

        axlow1.set_title('{} lowest frequency IMFs'.format(len(imfs) - highNums[0] - 1), fontsize = 16)
        axlow2.set_title('{} lowest frequency IMFs'.format(len(imfs) - highNums[1] - 1), fontsize = 16)
        axresid.set_title('Residual', fontsize = 16)

        axresid.plot(plotArgs[0], plotArgs[6], color = '#000000')
        axL.append(axresid)
    
    for ax in axL:
        ax.tick_params(axis = 'x', labelsize = 12)
        ax.tick_params(axis = 'y', labelsize = 12)

    plt.tight_layout()

    if show:
        plt.show()

    return highData1, highData2, lowData1, lowData2


def plot_imfs_noise(imfs, guessType = 'high', noiseNum = 2, collage = False, show = True):
    '''
    Plots assumed noise from IMF summation in a histogram, with overlaid graphs of fit probability distributions to check if assumption can be validated
    
    @param imfs: Input array of IMFs to be summed
    @param guessType: String of noise guess type ('high' or 'low' are possibilities)
    @param noiseNum: Number of IMFs to sum
    @param collage: Boolean that can optionally return certain plot parameters for external usage
    @param show: Boolean to show plot immediately after plot creation

    @return Array of plotted noise
    '''

    #! Need to fix K-S test
    
    if type(guessType) == str:
        if guessType.lower() == 'high':
            highFreq = True
        elif guessType.lower() == 'low':
            highFreq = False
        else:
            print('Invalid string argument for guessType')
            return
    else:
        ratio = guessType / len(imfs[0])
        if ratio >= 1. / 10:
            highFreq = True
        else:
            highFreq = False

    if highFreq:
        imfNoise = np.sum(imfs[: noiseNum], axis = 0)
    else:
        imfNoise = np.sum(imfs[-(noiseNum +1) : -1], axis = 0)


    mu  = np.mean(imfNoise)
    sig = np.std(imfNoise)
    

    plotArgs = imfNoise,
    if collage:
        return plotArgs

    plt.figure(figsize = (12, 6.75))
    n, bins, patches = plt.hist(imfNoise, bins = 250, normed = 1, color = '#CC33FF', label = 'Data bins')

    #? if p large (~ 1) cannot reject null; if p << 1, most likely can reject null (different distributions)
    #? kstest does not seem to be working as expected; alternative: i still dont understand what it is trying to do

    gauss   = stats.norm.pdf(bins, loc = mu, scale = sig)
    gauss_p = stats.kstest(imfNoise, 'norm', args = (mu, sig))

    tmp = '{:.2f}'.format(mu)
    if (tmp == '0.00') or (tmp == '-0.00'):
        muFmt = '{:.2}'.format(mu)
    else:
        muFmt = tmp

    tmp = '{:.2f}'.format(sig)
    if (tmp == '0.00') or (tmp == '-0.00'):
        sigFmt = '{:.2}'.format(sig)
    else:
        sigFmt = tmp

    plt.plot(bins, gauss, color = '#0066FF', linestyle = '--', label = r'Normal (fit): $\mu$ = {0}, $\sigma$ = {1}'.format(muFmt, sigFmt, gauss_p[1]))

    cGamma    = np.max(stats.cauchy.pdf(bins) / np.max(n))
    cauchyS   = stats.cauchy.pdf(bins, loc = 0, scale = cGamma)
    cauchyS_p = stats.kstest(imfNoise, 'cauchy', args = (0, cGamma))

    tmp = '{:.2f}'.format(cGamma)
    if (tmp == '0.00') or (tmp == '-0.00'):
        gamFmt = '{:.2}'.format(cGamma)
    else:
        gamFmt = tmp

    plt.plot(bins, cauchyS, color = '#FF6600', linestyle = '--', label = r'Cauchy (fit): $\gamma$ = {0}'.format(gamFmt, cauchyS_p[1]))

    plt.xlabel('Noise values')
    plt.ylabel('Probability (density)')
    plt.title('IMF noise ({})'.format(guessType))
    plt.legend(fontsize = 8)
    
    if show:
        plt.show()

    return imfNoise
    

#%# wrappers #%#

def run_plotImfs(inData, imfs = None, nbsym = False, toPlot = [], mainTitle = 'IMFs', show = True, figsize=(12,10)):
    '''
    Wrapper for plot_imfs

    @param inData: Input data for plotting
    @param imfs: Input array of IMFs for plotting
    @param nbsym: Boolean that would add extra data points near boundaries when calculating; breaks some datasets unless False
    @param toPlot: List of which IMFs to plot (default is all)
    @param show: Boolean to show plot immediately after plot creation
    '''
    real = np.isfinite(inData)
    if imfs is None:
        imfs = calc_imfs(inData[real], nbsym = nbsym)

    plot_imfs(inData[real], imfs, toPlot = toPlot, mainTitle = mainTitle, show = show, figsize = figsize)
    return imfs

def run_plotImfsSplit(inData, imfs = None, nbsym = False, highNum = 2, residual = False, mainTitle = 'Raw data', collage = False, show = True):
    # simple subplot-grid imf frequency split (high versus low) plot
    '''
    Wrapper for plot_imfs_split

    @param inData: Input data for plotting
    @param imfs: Input array of IMFs for summing and then plotting
    @param nbsym: Boolean that would add extra data points near boundaries when calculating; breaks some datasets unless False
    @param highNum: Number of high frequency IMFs to sum, starting from IMF1 (indexed at 0)
    @param residual: Boolean that optionally includes the residual function when summing low frequency IMFs (default False to disinclude residual)
    @param mainTitle: Title string of plot
    @param collage: Boolean that can optionally return certain plot parameters for external usage
    @param show: Boolean to show plot immediately after plot creation

    @return Tuple of HF summed data array and LF summed data array
    '''
    real = np.isfinite(inData)
    if imfs is None:
        imfs = calc_imfs(inData[real], nbsym = nbsym)

    outSplit = plot_imfs_split(inData[real], imfs, highNum = highNum, residual = residual, mainTitle = mainTitle, collage = collage, show = show)
    return outSplit


def run_plotImfsSplitComp(inData, imfs = None, nbsym = False, highNums = [2, 3], residual = False, plotRaw = True, mainTitle = 'Raw data', collage = False, show = True):
    '''
    Wrappper for plot_imfs_split_comp
    @param inData: Input data for plotting
    @param imfs: Input array of IMFs for summing and then plotting
    @param nbsym: Boolean that would add extra data points near boundaries when calculating; breaks some datasets unless False
    @param highNums: Number of high frequency IMFs to sum and compare
    @param residual: Boolean that optionally includes the residual function when summing low frequency IMFs (default False to disinclude residual)
    @param plotRaw: Boolean to optionally disinclude raw data plot above IMF summation comparison
    @param mainTitle: Title string of plot
    @param collage: Boolean that can optionally return certain plot parameters for external usage
    @param show: Boolean to show plot immediately after plot creation

    @return Tuple of both HF summed data arrays and borh LF summed data arrays
    '''
    real = np.isfinite(inData)
    if imfs is None:
        imfs = calc_imfs(inData[real], nbsym = nbsym)
        
    outSplit = plot_imfs_split_comp(inData[real], imfs, highNums = highNums, residual = residual, plotRaw = plotRaw, mainTitle = mainTitle, collage = collage, show = show)
    return outSplit

def run_plotImfsNoise(inData, imfs = None, nbsym = False, noiseNum = 2, guessType = 'high', show = True):
    '''
    Wrapper for plot_imfs_noise
    @param inData: Input data for plotting
    @param imfs: Input array of IMFs to be summed
    @param nbsym: Boolean that would add extra data points near boundaries when calculating; breaks some datasets unless False
    @param guessType: String of noise guess type ('high' or 'low' are possibilities)
    @param noiseNum: Number of IMFs to sum
    @param collage: Boolean that can optionally return certain plot parameters for external usage
    @param show: Boolean to show plot immediately after plot creation

    @return Array of noise data values
    '''
    real = np.isfinite(inData)
    if imfs is None:
        imfs = calc_imfs(inData[real], nbsym = nbsym)

    outNoise = plot_imfs_noise(imfs, guessType = guessType, noiseNum = noiseNum, show = show)
    return outNoise

def run_plotImfsSplitNoise(inData, imfs = None, nbsym = False, highNum = 2, residual = False, mainTitle = 'Raw data', noiseNum = 2, guessType = 'high', show = False):
    '''
    Wrapper for both plot_imfs_split and plot_imfs_noise

    @param inData: Input data for plotting
    @param imfs: Input array of IMFs for summing and then plotting
    @param nbsym: Boolean that would add extra data points near boundaries when calculating; breaks some datasets unless False
    @param highNum: Number of high frequency IMFs to sum
    @param residual: Boolean that optionally includes the residual function when summing low frequency IMFs (default False to disinclude residual)
    @param mainTitle: Title string of plot
    @param guessType: String of noise guess type ('high' or 'low' are possibilities)
    @param noiseNum: Number of IMFs to sum
    @param show: Boolean to show plot immediately after plot creation

    @return Tuple of split tuple and noise array

    '''
    real = np.isfinite(inData)
    if imfs is None:
        imfs = calc_imfs(inData[real], nbsym = nbsym)

    outSplit = plot_imfs_split(inData[real], imfs, highNum = highNum, residual = residual, mainTitle = mainTitle, show = show)
    outNoise = plot_imfs_noise(imfs, guessType = guessType, noiseNum = noiseNum, show = show)

    plt.show()

    return outSplit, outNoise



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import collections

import numpy as np
import scipy.optimize as optimize
import skdaccess.utilities.pbo_util as pbo_utils
from skdiscovery.data_structure.framework import PipelineItem

import skdiscovery.utilities.patterns.pbo_tools as pbo_tools


class Mogi_Inversion(PipelineItem):
    '''
    Perform a mogi source inversion on a set of gps table data

    The source is assumed to be a mogi source (point source), but other source models can be selected.
    Assumes directions are named ('dN', 'dE', 'dU').
    '''
    def __init__(self, str_description, ap_paramList, pca_name, column_names=['dN', 'dE', 'dU']):
        '''
        Initialize Mogi analysis item

        @param str_description: Description of item
        @param ap_paramList[source_type]: Type of magma chamber source model to use (default-mogi,finite_sphere,closed_pipe,constant_open_pipe,rising_open_pipe,sill)
        @param pca_name: Name of pca result
        @param column_names: The data direction column names
        ''' 

        self.pca_name = pca_name
        self.column_names = column_names
        super(Mogi_Inversion, self).__init__(str_description, ap_paramList)
        self.ap_paramNames = ['source_type']
        
        
    def FitPCA(self, hPCA_Proj):
        '''
        Determine the timing of the inflation event from the first component of the pca projection

        fits A * arctan( (t - t0) / c ) + B to the first pca projection, in order to estimate
        source amplitude parameters

        @param hPCA_Proj: The sklearn PCA 

        @return ct: the t0, c, and B parameters from the fit
        @return pA[0]: the fitted amplitude parameter
        '''

        fitfunc = lambda p,t: p[0]*np.arctan((t-p[1])/p[2])+p[3]
        errfunc = lambda p,x,y: fitfunc(p,x) - y
    
        dLen = len(hPCA_Proj[:,0])
        pA, success = optimize.leastsq(errfunc,[1.,dLen/2.,1.,0.],args=(np.arange(dLen),hPCA_Proj[:,0]))
        ct = pA[1:3]

        return ct, pA[0]
        

    def FitTimeSeries(self, pd_series, ct):
        '''
        Fits the amplitude and offset of an inflation event given the time and length of the event
        
        Fits A and B in A * arctan( (t - t0) / c) + B
        
        @param pd_series: Time series to be fit
        @param ct: the time constants for the arctan

        @return res: Amplitude of the fit
        @return perr_leastsq: Error of the fit
        '''

        fitfunc2 = lambda p,c,t: p[0]*np.arctan((t-c[0])/c[1])+p[1]
        errfunc2 = lambda p,c,x,y: fitfunc2(p,c,x) - y

        dLen = len(pd_series)
        
        pA, pcov = optimize.leastsq(errfunc2,[1.,0.],args=(ct,np.arange(dLen),pd_series))
        # res = fitfunc2(pA,ct,np.arange(dLen))[-1]-fitfunc2(pA,ct,np.arange(dLen))[0]
        res = pA[0]*np.pi
        
        s_sq = (errfunc2(pA,ct,np.arange(dLen),pd_series)**2).sum()/(len(pd_series)-2)
        pcov = pcov * s_sq
        error = [] 
        for i in range(len(pA)):
            try:
              error.append(np.absolute(pcov[i][i])**0.5)
            except:
              error.append( 0.00 )
        perr_leastsq = np.array(error) 

        return res, perr_leastsq


    def process(self, obj_data):
        '''
        Finds the magma source (default-mogi) from PBO GPS data.

        Assumes time series columns are named ('dN', 'dE', 'dU'). Predicts the
        location of the magma source using scipy.optimize.curve_fit

        The result is added to the data wrapper as a list, with the four
        elements describing the location of the magma source:
        res[0] = latitude
        res[1] = longitude
        res[2] = source depth (km)
        res[3] = volume change (meters^3)

        @param obj_data: 
        '''
        h_pca_name = self.pca_name
        
        exN = {'mogi':0,'finite_sphere':1,'closed_pipe':1,'constant_open_pipe':1,'rising_open_pipe':2,'sill':0}
        try:
            mag_source = getattr(pbo_tools,self.ap_paramList[0]().lower())
            ExScParams = tuple(np.ones((exN[self.ap_paramList[0]().lower()],)))
        except:
            mag_source = pbo_tools.mogi
            ExScParams = ()
            print('No source type called '+self.ap_paramList[0]()+', defaulting to a Mogi source.')
                
        projection = obj_data.getResults()[h_pca_name]['Projection']
        start_date = obj_data.getResults()[h_pca_name]['start_date']
        end_date = obj_data.getResults()[h_pca_name]['end_date']        

        ct, pca_amp = self.FitPCA(projection)
        pca_amp *= np.pi

        xvs = []
        yvs = []
        zvs = []

        label_list = []

        for label, data in obj_data.getIterator():
            label_list.append(label)
            for column in self.column_names:
                distance,f_error = self.FitTimeSeries(data.loc[start_date:end_date,column], ct)
                if column == self.column_names[1]:
                    xvs.append(distance)
                elif column == self.column_names[0]:
                    yvs.append(distance)
                elif column == self.column_names[2]:
                    zvs.append(distance)
                else:
                    print('Ignoring column: ', label)

        xvs = np.array(xvs)*1e-6
        yvs = np.array(yvs)*1e-6
        zvs = np.array(zvs)*1e-6

        ydata = np.hstack((xvs, yvs,zvs)).T
        station_list = obj_data.get().keys()
        meta_data = obj_data.info()
        station_coords = pbo_utils.getStationCoords(meta_data, station_list)
        
        dimensions = ('x','y','z')
        xdata = []
        for dim in dimensions:
            for coord in station_coords:
                xdata.append((dim, coord[0], coord[1]))

        coord_range = np.array(pbo_utils.getLatLonRange(meta_data, station_list))

        lat_guess = np.mean(coord_range[0,:])
        lon_guess = np.mean(coord_range[1,:])

        fit = optimize.curve_fit(mag_source, xdata, ydata, (lat_guess, lon_guess, 5, 1e-4)+ExScParams)

        res = collections.OrderedDict()

        res['lat'] = fit[0][0]
        res['lon'] = fit[0][1]
        res['depth'] = fit[0][2]
        res['amplitude'] = fit[0][3]
        res['labels'] = label_list
        if len(fit[0])>4:
            res['ex_params'] = fit[0][4:]
        else:
            res['ex_params'] = np.nan
        res['pca_amplitude'] = pca_amp
        res['source_type'] = self.ap_paramList[0]().lower()

        obj_data.addResult(self.str_description, res)

        # lat_fit_range = (np.min(lat_list)-0.15, np.max(lat_list)+0.15)
        # lon_fit_range = (np.min(lon_list)-0.15, np.max(lon_list)+0.15)

        

        # res = optimize.brute(self.mogi, (lat_fit_range, lon_fit_range,
        #                                  (1,10), (1e-5, 1e-3)),
        #                      args = (xvs*1e-6, yvs*1e-6, zvs*1e-6,
        #                              station_list, meta_data))


def MogiVectors(mogi_res,station_lat_list,station_lon_list,flag3D=False):
    '''
    Creates a set of mogi vectors for plotting

    @param mogi_res: Magma source
    @param station_lat_list: List of station latitudes
    @param station_lon_list: List of station longitudes
    @param flag3D: Make vectors 3 dimensional, defaults to False (2D)
    '''
    
    # grab the correct magma forward function
    mag_source = getattr(pbo_tools,mogi_res['source_type'])

    mogi_x_disp = []
    mogi_y_disp = []
    for lat, lon in zip(station_lat_list, station_lon_list):
        mogi_data = []

        mogi_data.append( ('x', lat, lon) )
        mogi_data.append( ('y', lat, lon) )
        if flag3D:
            mogi_data.append( ('z', lat, lon) )

        if np.isnan(mogi_res['ex_params']).any()==True:
            res = mag_source(mogi_data, mogi_res['lat'], mogi_res['lon'],
                             mogi_res['depth'], mogi_res['amplitude'])
        elif len(mogi_res['ex_params'])==1:
            res = mag_source(mogi_data, mogi_res['lat'], mogi_res['lon'],
                             mogi_res['depth'], mogi_res['amplitude'],mogi_res['ex_params'][0])
        elif len(mogi_res['ex_params'])==2:
            res = mag_source(mogi_data, mogi_res['lat'], mogi_res['lon'],
                             mogi_res['depth'], mogi_res['amplitude'],mogi_res['ex_params'][0],mogi_res['ex_params'][1])

        mogi_x_disp.append(res[0])
        mogi_y_disp.append(res[1])

    # the scaling here is because mogi distance vectors are in km
    # the factor converts to mm and scales by the pca change
    mogi_x_disp = np.array(mogi_x_disp) * 1e6 / (mogi_res['pca_amplitude'])
    mogi_y_disp = np.array(mogi_y_disp) * 1e6 / (mogi_res['pca_amplitude'])
    
    return mogi_x_disp, mogi_y_disp



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Soubhik Barari
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


"""Provides some example random walk functions in the expected format.

All random walk functions have the following signature:

(pos: tuple of input point, grid: bounds for walk, step_size: maximal step size) -> (output pos: tuple)

"""

import  random
from    numpy.random import multivariate_normal


def uniform_walk(pos, grid, step_size=None):
    '''
    A uniform random walk function
    
    @param pos: tuple of input point
    @param grid: bounds for walk
    @param step_size: maximal step size
    
    @return position tuple
    '''
    newpos = [float(random.randrange(grid[0]-grid[2]/2, grid[0]+grid[2]/2)), 
            float(random.randrange(grid[1]-grid[3]/2, grid[1]+grid[3]/2))]
            
    return newpos


def gaussian_walk(pos, grid, step_size=None):
    '''
    A gaussian random walk function
    
    @param pos: tuple of input point
    @param grid: bounds for walk
    @param step_size: maximal step size
    
    @return position tuple
    '''

    step = multivariate_normal(mean=[0,0], cov=[[grid[0],0],[0,grid[1]]], size=1)[0]
    newpos = [pos[0] + step[0], pos[1] + step[1]]       

    newpos = keep_in_bound(newpos, grid)
    return newpos


def keep_in_bound(pos, grid):
    '''
    Function for truncating and bounding the random walk to within the defined grid
    
    @param pos: tuple of the point to be checked
    @param grid: the bounds for limiting the walk
    
    @return position tuple after bounding the point
    '''
    pos[0] = max(pos[0], grid[0]-grid[2]/2)
    pos[0] = min(pos[0], grid[0]+grid[2]/2)
    pos[1] = max(pos[1], grid[1]-grid[3]/2)
    pos[1] = min(pos[1], grid[1]+grid[3]/2)
    return pos





from .dataremover import DataRemover
from .hyperbolictan import HTanFilter
from .interpolate import InterpolateFilter
from .kalman import KalmanFilter
from .lowpass import LowPassFilter
from .median import MedianFilter
from .trend import TrendFilter
from .offset_detrend import OffsetDetrend


__all__ = ['DataRemover', 'HTanFilter', 'InterpolateFilter',
           'KalmanFilter', 'LowPassFilter', 'MedianFilter',
           'TrendFilter','OffsetDetrend']



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import random

class StageContainer:
    ''' Container to hold a stage for the DiscoveryPipeline. '''
    
    def __init__(self, obj_content, obj_runmethod = None, obj_perturbmethod = None, obj_reset = None):
        '''
        Get the object and its run method into this conainer.
        
        @param obj_content: filter, analysis, or accumlator
        @param obj_runmethod: Run method of the obj_content (default process)
        @param obj_perturbmethod: Perturb method of the obj_content (default peturbParams)
        @param obj_reset: Reset method of the obj_content (default resetParams)
        '''
        self.obj_content    = obj_content

        if obj_runmethod != None:
            self.runmethod      = obj_runmethod
        else:
            self.runmethod = self.obj_content.process

        if obj_perturbmethod != None:
            self.perturbmethod  = obj_perturbmethod
        else:
            self.perturbmethod  = self.obj_content.perturbParams


        if obj_reset != None:
            self.resetmethod  = obj_reset
        else:
            self.resetmethod  = self.obj_content.resetParams

        
    def run(self, obj_data_container):
        '''
        Execute the obj_content run method

        @param obj_data_container: Data container to be passed to the held obj_content's run method
        '''
        self.runmethod(obj_data_container)

    def perturb(self):
        ''' Execute the obj_content peturb method '''
        self.perturbmethod()

    def reset(self):
        ''' Execute the obj_content reset method '''
        self.resetmethod()

    def getMetadata(self):
        '''
        Retrieves the obj_content metadata
        
        @return obj_content metadata
        '''
        return self.obj_content.getMetadata()

    def getObjects(self):
        '''
        Return the obj_content in a list 

        @return Contained object in a list
        '''
        return [self.obj_content]

    #+++++++ NEW Victor ++++++++
    def getMetadataType(self):
        '''
        Get metadata type

        @return String: container type
        '''
        return "SC"


    #+++++++ NEW Victor ++++++++
    def getMetadataNestedTypes(self):
        '''
        Get the metadata along with container type

        @return string of container and metadata
        '''
        returnString= "#SC("+ self.getMetadata() +")"        
        return returnString 

    #+++++++ NEW Victor ++++++++
    def getMetadataNestedGraph(self):
        '''
        Get the nested graph for the container

        @return String: Stage container subgraph
        '''
        returnString= "\n subgraph cluster { \n \""+ self.getMetadata() +"\"; \n label=\"SC\" }\n"        
        return returnString 

    
class StageContainerAlternative:
    ''' 
    Stage Container that holds a list of stage containers and randomly chooses one to use.
    '''
    
    currentContainer =[]
    
    def __init__(self, list_stagecontainers):
        '''
        Initialize the StageContainerAlternative

        @param list_stagecontainers: List of stage containers
        '''
        
        self.list_stagecontainers = list_stagecontainers
        self.currentContainer = self.list_stagecontainers[0]
#        self.currentContainer = random.choice(self.list_stagecontainers)
#        self.perturb()
        
    def run(self, obj_data_container):
        ''' 
        Run the currently selected stage container.

        @param obj_data_container: Data container to be passed to the current stagecontainer
        '''
        self.currentContainer.run(obj_data_container)

    def perturb(self):
        ''' choose one of the containers as an alternative and perturb its parameters'''
        self.currentContainer = random.choice(self.list_stagecontainers)
        self.currentContainer.perturb()

    def getMetadata(self):
        ''' 
        Return metadata from the current container 

        @return metadata from the currently selected container
        '''
        return self.currentContainer.getMetadata()

    def getObjects(self):
        '''
        retrieve the current container as a list

        @return Current container being used as a list
        '''

        return self.currentContainer.getObjects()

    def reset(self):
        '''
        Reset the current chosen StageContainer

        self.currentContainer.reset()
        '''

    #++++++++ NEW Victor +++++++++++++++++
    def getMetadataType(self):
        '''
        Get metadata type

        @return String: container type
        '''
        return "SC_Alternative"

    #++++++++ NEW Victor +++++++++++++++++
    def getMetadataNestedTypes(self):
        '''
        Get the metadata along with container type

        @return string of container and metadata
        '''
        
        returnString="#ALT("
        for e in self.list_stagecontainers:
                returnString = returnString + e.getMetadataNestedTypes()
        returnString=returnString + ")"
        
        return returnString 

    #++++++++ NEW Victor +++++++++++++++++
    def getMetadataNestedGraph(self):
        '''
        Get the nested graph for the container

        @return String: Container subgraph
        '''
        
        returnString="\n subgraph cluster { \n"
        for e in self.list_stagecontainers:
                returnString = returnString + "\"" + e.getMetadataNestedTypes() +"\"; \n"
        returnString=returnString + " label=\"ALT\" } \n"
        
        return returnString 
        
        


class StageContainerIncrementalAdd:
    ''' In each perturb call, it incrementally adds one of the filters specified in the constructor. '''
    

    def __init__(self, list_stagecontainers):
        '''
        Initialize the container

        @param list_stagecontainers: List of stage containers.
        '''
        self.length = len(list_stagecontainers)
        self.list_AllStagecontainers = list_stagecontainers

        self.currentindex = 0
        self.list_currentContainers      = []

        self.perturb()
        
    def reset(self):
        ''' Reset the container so that it will only run the first stage container again. '''
        self.list_currentContainers     = []
        self.currentindex =0
        self.perturb()
    
    def run(self, obj_data_container):
        ''' Run the current list of stage containers'''
        for c in self.list_currentContainers:
            c.run(obj_data_container)

    def perturb(self):
        ''' Add another stage container to the current list of stage containers '''
        if self.currentindex < self.length:
            self.list_currentContainers.append(self.list_AllStagecontainers[self.currentindex])
            self.currentindex+=1
        
        # for c in self.list_currentContainers:
        #     c.perturb()
            
    def getMetadata(self):
        '''
        Return the metadata from the currently used stage containers.
        
        @return List of metadata from current containers
        '''
        metadataList = [s.getMetadata() for s in self.list_currentContainers]
        return metadataList

    def getObjects(self):
        '''
        Retrieve objects in the current list of stage containers

        @return List of current obj_content from the current list of stage containers
        '''
        obj_list = []
        for container in self.list_currentContainers:
            for obj in container.getObjects():
                obj_list.append(obj)

        return obj_list

    #++++++++ NEW Victor +++++++++++++++++
    def getMetadataType(self):
        '''
        Get metadata type

        @return String: container type
        '''        
        
        return "SC_IncrementalAdd"
    
    #++++++++ NEW Victor +++++++++++++++++
    def getMetadataNestedTypes(self):
        '''
        Get the metadata along with container type

        @return string of container and metadata
        '''
        returnString="#INCA("
        for e in self.list_stagecontainers:
                returnString = returnString + e.getMetadataNestedTypes()
        returnString=returnString + ")"

    #++++++++ NEW Victor +++++++++++++++++
    def getMetadataNestedGraph(self):
        '''
        Get the nested graph for the container

        @return String: Container subgraph
        '''
        
        returnString="\n subgraph cluster { \n"
        for e in self.list_stagecontainers:
                returnString = returnString + "\""+e.getMetadataNestedTypes()+"\"; \n"
        returnString=returnString + "\n label=\"INCA\" } \n"
        
        return returnString 



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem
import pandas as pd
import numpy as np

class DataRemover(PipelineItem):
    ''' Sets specified series data to NaN '''
    
    def __init__(self, str_description, start=None, end=None, labels=None, column_names = None):
        ''' 
        Initialize DataRemover
        
        @param str_description: String describing filter
        @param start: Starting index value
        @param end: Ending index value (inclusive)
        @param labels: List of labels used to select data to be removed (None will operate on all labels)
        @param column_names: List of column names to select data to be removed (None will operate on all columns)
        '''
        self.labels = labels
        self.column_names = column_names
        self.start = start
        self.end = end

        super(DataRemover, self).__init__(str_description, [])

    def process(self, obj_data):
        ''' 
        NaN's data from DataWrapper

        @param obj_data: Input DataWrapper, which will be modified in place
        '''

        labels = self.labels
        column_names = self.column_names
        
        for label, data, err in obj_data.getIterator():
            if (labels is None or label in labels) and \
               (column_names is None or data.name in column_names):

                index = data.index

                if self.start is None:
                    start = index[0]
                else:
                    start = self.start

                if self.end is None:
                    end = index[-1]
                else:
                    end = self.end

                new_nans = np.empty(len(data[start:end]))
                new_nans[:] = np.NAN
                new_nans = pd.Series(new_nans, index=data[start:end].index)

                data.loc[start:end] = new_nans



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework.base import PipelineItem
from skdaccess.utilities.grace_util import computeEWD
import pandas as pd


class CalibrateGRACE(PipelineItem):
    '''
    Calibrate Grace Data

    Averages the three solutions and applies a scale factor
    '''
    
    def __init__(self, str_description, ewd_column_name='EWD', round_dates = True):
        '''
        Initialize GRACE calibration filter

        @param str_description: String describing filter
        @param ewd_column_name: Name of new column for the calibrated GRACE data
        @param round_dates: Option for rounding to dates to the nearest day
        '''
        self.ewd_column_name = ewd_column_name
        self.round_dates = round_dates
        super(CalibrateGRACE, self).__init__(str_description)
        
    def process(self, obj_data):
        '''
        Calibrates GRACE, updating in place

        @param obj_data: Table data wrapper
        '''
        label_list = []
        frame_list = []
        for label, data in obj_data.getIterator():
            frame_list.append(pd.DataFrame(computeEWD(data, obj_data.info(label)['scale_factor'],
                                          round_nearest_day=self.round_dates), columns=[self.ewd_column_name]))

            frame_list[-1].loc[:,self.ewd_column_name + '_Error'] = obj_data.info(label)['measurement_error']
            label_list.append(label)
            
        obj_data.updateFrames(label_list,frame_list)
        obj_data.default_columns = [self.ewd_column_name]



# The MIT License (MIT)
# Copyright (c) 2018 Massachusetts Institute of Technology
#
# Authors: Justin Li, Juha Vierinen, Bill Rideout, and Shun-Rong Zhang
#
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and
# NASA AIST Project "Computer-Aided Discovery of Earth Surface
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import glob
import time
import warnings
import os
import traceback
import sys
from optparse import OptionParser


# third party imports
from mpl_toolkits.basemap import Basemap
import numpy
import matplotlib.pyplot as plt
import scipy.interpolate
import scipy.signal
import pandas as pd
import numpy as np
import tqdm
from scipy.stats import ks_2samp
from scipy.stats import anderson_ksamp

# Framework imports
from skdiscovery.utilities.planetary.map_util import wgs84_distance

# for calculating geographical distance
def geocalc(lat1, lon1, lat2, lon2):

    return wgs84_distance((lat1, lon1), (lat2, lon2))


def get_lp_tec(tvec, vtec_est, window_length=481, polyorder=3):
    """get_lp_tec returns a low pass version of the vertical tec at the same time spacing
    as vtec_est (that is, at the times given by tvec).  If problem, returns None.  Where data
    cannot be low pass filtered, returns numpy.nan values

    Inputs:
        tvec - input time array in float days
        vtec_est - input vertical tec arr, len = len(tvec)
        window_length - number of 15 second intervals to window over. Default is 481 (2 hours)
            Must be odd
        polyorder - order of polynomial fit to window. Default is 3.
    """
    gap_size = 1.0 / (24*4) # tvec is in float days, so gap_size is 15 minutes
    reg_step = 1.0 / (24*60*4) # step size = 15 seconds in float days
    if window_length % 2 == 0:
        raise ValueError('Window len must be odd, not %i' % (window_length))

    #ret_times = None
    ret_values = None # the numpy values to return

    # first task - break line into gap free sections
    indices = [0]

    for i in range(1, len(tvec)):
        if tvec[i] - tvec[i-1] > gap_size:
            indices.append(i)
    indices.append(len(tvec))

    # deal with each continuous segment separately
    for i in range(len(indices) - 1):
        sub_tvec = tvec[indices[i]:indices[i+1]]
        if len(sub_tvec) < 10:
            lp_values = numpy.zeros((len(sub_tvec),), dtype=numpy.float)
            lp_values[:] = numpy.nan
            if ret_values is None:
                ret_values = lp_values
            else:
                ret_values = numpy.concatenate((ret_values, lp_values))
            continue
        sub_vtec_est = vtec_est[indices[i]:indices[i+1]]
        # create regular time array covering the same span
        reg_tvec = numpy.arange(sub_tvec[0], sub_tvec[-1], reg_step)

        reg_func = scipy.interpolate.interp1d(sub_tvec, sub_vtec_est)
        reg_values = reg_func(reg_tvec)

        # now apply window filter
        if window_length > len(reg_values):
            this_window = len(reg_values)
            if this_window % 2 == 0:
                this_window -= 1
        else:
            this_window = window_length
        lp_reg_values = scipy.signal.savgol_filter(reg_values, this_window, polyorder)

        # now interp windowed data
        lp_func = scipy.interpolate.interp1d(reg_tvec, lp_reg_values)
        for i in range(10):
            try:
                lp_values = lp_func(sub_tvec[:-1 + (-1*i)])
                break
            except ValueError:
                # we may be beyond the interpolation area, drop last point
                continue

        # set values to nan if window was not long enough
        if window_length > len(reg_values):
            lp_values[:] = numpy.nan

        # pad with nan if needed
        if len(lp_values) < len(sub_tvec):
            pad_arr = numpy.zeros((len(sub_tvec)-len(lp_values),), dtype=numpy.float)
            pad_arr[:] = numpy.nan
            lp_values = numpy.concatenate((lp_values, pad_arr))

        if ret_values is None:
            ret_values = lp_values
        else:
            ret_values = numpy.concatenate((ret_values, lp_values))

    if ret_values is None:
        return(None)

    # make sure the length is right by padding begining and end with nan
    while(len(ret_values) < len(tvec)):
        ret_values = numpy.insert(ret_values, 0, numpy.nan)
        if len(ret_values) >= len(tvec):
            break
        ret_values = numpy.insert(ret_values, len(ret_values), numpy.nan)

    return(ret_values)





# --------------------------------------------------------------------------------
# --------------------------------------------------------------------------------
# a bunch of my functions
# --------------------------------------------------------------------------------
# --------------------------------------------------------------------------------

def getRawStitch(DOYs,llat,ulat,llon,rlon,year=2016):
    anasite = []
    anarray = []
    for adoy in DOYs:
        sites,tecs = getRawTEC(adoy,llat,ulat,llon,rlon,year)
        anasite.append(sites)
        anarray.append(tecs)

    catsite = anasite[-1].copy()
    for ii in np.arange(len(anasite)-1):
        appsites = []
        for asite in anasite[ii].index:
            appsites.append(asite)
        catsite.append(anasite[ii].loc[appsites])

    return catsite,pd.concat(anarray)


def fixTECoffset(siteprnTEC,doyN,dchk=3,dcut=.25,mjump=1):
    indices = list(np.hstack((0,np.where((np.abs(np.diff(siteprnTEC.vtec.values))>dcut)==1)[0]+1,len(siteprnTEC)-1)))
    for kk in np.arange(-dchk,dchk+1):
        atmp = np.argmin(np.abs(siteprnTEC.index.values-doyN-kk))
        if atmp not in indices:
            indices.append(atmp)
    indices.sort()

    itemp = indices.copy()
    for ii in np.arange(1,len(itemp[1:-1])):
        aidx = indices[ii]
        if (siteprnTEC.index.values[aidx]-siteprnTEC.index.values[aidx-1])<(mjump/60/24):
            siteprnTEC.vtec.values[indices[ii-1]:indices[ii]] += siteprnTEC.vtec.values[indices[ii]] - siteprnTEC.vtec.values[indices[ii]-1]
            indices[ii] = indices[ii-1]

# # need to check in between days to update index?
# dayindx = []
# for kk in np.arange(-dchk,dchk+1):
#     atmp = np.argmin(np.abs(siteprnTEC.index.values-4-kk))
#     if atmp not in indices:
#         dayindx.append(atmp)
#         indices.append(atmp)

#     elif aidx in dayindx:
#         print(siteprnTEC.vtec.values[indices[ii]] - siteprnTEC.vtec.values[indices[ii]-1])
#         indices[ii] = indices[ii-1]


def findTECevents(rawdata,dayNum,hrEvent,pwin=200,nstd=10,thrstd=.75,verbose=False, fixOffset=False):
    if len(rawdata)==2:
        raw_1 = rawdata[0]
        raw_2 = rawdata[1]
        raw_3 = None
    if len(rawdata)==3:
        # will code actual use of this later maybe
        raw_3 = rawdata[2]

    resbuf = []
    for asite in tqdm.tqdm(np.sort(raw_2[0].index.values)):
        if asite in raw_1[0].index.values:
            sitedata = raw_2[1].iloc[(raw_2[1].site.values==asite),:]
            p_sitedata = raw_1[1].iloc[(raw_1[1].site.values==asite),:]
            p_prns = np.unique(p_sitedata.prn.values)
            for aprn in np.unique(sitedata.prn.values):
                if aprn not in p_prns:
                    if verbose:
                        print('Previous day missing PRN'+str(aprn))
                    pass
                else:
                    pairdat = sitedata.iloc[sitedata.prn.values==aprn,:]
                    p_pairdat = p_sitedata.iloc[p_sitedata.prn.values==aprn,:]
                    if fixOffset:
                        fixTECoffset(pairdat,1)
                        fixTECoffset(p_pairdat,1)
                    if np.argmin(np.abs((pairdat.index.values-(dayNum-1))*24-hrEvent))<=0 or np.argmin(np.abs((pairdat.index.values-(dayNum-1))*24-hrEvent)) >= len(pairdat) or np.argmin(np.abs((p_pairdat.index.values-(dayNum-2))*24-hrEvent))<=0 or np.argmin(np.abs((p_pairdat.index.values-(dayNum-2))*24-hrEvent)) >= len(p_pairdat):
                        pass
                    else:
                        dtec = pairdat.vtec.values - get_lp_tec((pairdat.index.values-(dayNum-1))*24,pairdat.vtec.values)
                        p_dtec = p_pairdat.vtec.values - get_lp_tec((p_pairdat.index.values-(dayNum-2))*24,p_pairdat.vtec.values)

                        tidx = np.argmin(np.abs((pairdat.index.values-(dayNum-1))*24-hrEvent))-int(pwin/2)
                        p_tidx = np.argmin(np.abs((p_pairdat.index.values-(dayNum-2))*24-hrEvent))-int(pwin/2)
                        # zero out nan's and make the other half zero too
                        dtec[np.isnan(dtec)] = 0
                        p_dtec[np.isnan(p_dtec)] = 0
                        # okay to do by multiply because same fs and time aligned via tidx
                        dayLen = len(dtec[tidx:tidx+pwin]); p_dLen = len(p_dtec[p_tidx:p_tidx+pwin])
                        if dayLen > p_dLen:
                            p_dtec = np.hstack((p_dtec,np.zeros(dayLen-p_dLen)))
                        elif p_dLen > dayLen:
                            dtec = np.hstack((dtec,np.zeros(p_dLen-dayLen)))

                        try:
                            nanmask = (dtec[tidx:tidx+pwin]!=0)*(p_dtec[p_tidx:p_tidx+pwin]!=0)
                            dtec[tidx:tidx+pwin] *= nanmask
                            p_dtec[p_tidx:p_tidx+pwin] *= nanmask
                            siglvl = anderson_ksamp([dtec[tidx:tidx+pwin],p_dtec[p_tidx:p_tidx+pwin]]).significance_level
                            pval = ks_2samp(dtec[tidx:tidx+pwin],p_dtec[p_tidx:p_tidx+pwin]).pvalue
                        except:
                            siglvl = 1; pval = 1

                        if siglvl == 1 and pval == 1:
                            if verbose: print('No Matching Data')
                            pass
                        elif np.sum(pairdat.iloc[tidx:tidx+pwin].vtec.rolling(nstd).std().values>thrstd)>nstd or np.sum(p_pairdat.iloc[p_tidx:p_tidx+pwin].vtec.rolling(nstd).std().values>thrstd)>nstd:
                            if verbose: print('Too Noisy for PRN'+str(aprn))
                            pass
                        elif np.abs((pairdat.iloc[max([tidx,0])].name-(dayNum-1))*24-(p_pairdat.iloc[max([p_tidx,0])].name-(dayNum-2))*24)>.5:
                            # if start is off by more than 1/2 hour, pass
                            if verbose: print('Not Time Aligned')
                            pass
                        elif (pairdat.iloc[min([tidx+pwin,len(pairdat)-1])].name-(dayNum-1))*24<hrEvent:
                            # if day of event, data ends before the event, pass
                            if verbose: print('Not Within the Event Time')
                            pass
                        elif np.max(np.diff((pairdat.iloc[tidx:tidx+pwin].index.values-(dayNum-1))*24))>.25 or np.max(np.diff((p_pairdat.iloc[p_tidx:p_tidx+pwin].index.values-(dayNum-2))*24))>.25:
                            # if big (>15min) gap in middle of chunk (ie if because start after event)
                            if verbose:print('Missing Data')
                            pass
                        elif siglvl<.05 and pval<.05:
                            if verbose: print(asite,aprn)
                            resbuf.append([(pairdat.index.values-(dayNum-1))*24,(p_pairdat.index.values-(dayNum-2))*24,dtec,p_dtec,
                                          asite,aprn])
        else:
            if verbose: print('Previous day missing '+asite)
            pass

    print('Found Events for '+str(len(resbuf))+' Site-PRN Pairs')
    return resbuf


def plotTECres(pidx,resbuf,hrEvent,pwin=200):
    tidx = np.argmin(np.abs(resbuf[pidx][0]-hrEvent))-int(pwin/2);
    p_tidx = np.argmin(np.abs(resbuf[pidx][1]-hrEvent))-int(pwin/2)
    plt.plot(resbuf[pidx][0][tidx:tidx+pwin],resbuf[pidx][2][tidx:tidx+pwin]);
    plt.plot(resbuf[pidx][1][p_tidx:p_tidx+pwin],resbuf[pidx][3][p_tidx:p_tidx+pwin]+.15);
    plt.title(resbuf[pidx][4]+'-'+str(resbuf[pidx][5]));
    plt.xlabel('Time (Hrs)'); plt.ylabel('dTEC (TEC Units)');



def makeMap(lat_0,lon_0,dbuffer=5,projection='gnom',resolution='i'):
    # broader region of event
    llat = lat_0-dbuffer
    ulat = lat_0+dbuffer
    llon = lon_0-dbuffer
    rlon = lon_0+dbuffer

    bmap = Basemap(projection=projection,lat_0=lat_0,lon_0=lon_0, urcrnrlat=ulat, urcrnrlon=rlon,
                   llcrnrlat=llat,llcrnrlon=llon, resolution=resolution)
    latlons = [llat,ulat,llon,rlon]

    return bmap, latlons


def findPRNs(raw_tec,eventHr,doyN,lat_0,lon_0,latWin=5,lonWin=5,nThreshold=1000):
    potentialPRN = []
    for aprn in np.arange(33):
        site_day = raw_tec[1].iloc[(raw_tec[1].prn.values==aprn),:]
        site_day = site_day.iloc[((site_day.index.values-(doyN-1))*24>eventHr)*((site_day.index.values-(doyN-1))*24<(eventHr+1)),:]
        site_day = site_day.iloc[(np.abs(site_day.pplat.values-lat_0)<latWin)*(np.abs(site_day.pplon.values-lon_0)<lonWin),:]
        if len(site_day)>nThreshold:
            potentialPRN.append(aprn)
    return potentialPRN


def genDTecs(aprn,raw_tec,doyN):
    site_day = raw_tec[1].iloc[(raw_tec[1].prn.values==aprn),:]
    dtecDat = site_day.assign(dtec=0)

    for asite in np.unique(site_day.site.values):
        prdat = site_day.iloc[site_day.site.values==asite]
        dtec = prdat.vtec - get_lp_tec((prdat.index.values-(doyN-1))*24,prdat.vtec.values)
        dtecDat.loc[dtecDat.site.values==asite,'dtec'] = dtec
    return dtecDat


def plotPRNd(raw_tec,dtecDat,eventHr,doyN,lat_0,lon_0,m,fsize=(10,10),clim=.1,ms=5):
    plt.figure(figsize=fsize);m.drawcoastlines();m.plot(lon_0,lat_0,'r^',latlon=True,markersize=15);
    m.plot(raw_tec[0].rec_lon.values,raw_tec[0].rec_lat.values,'bs',latlon=True)
    site_day = dtecDat.iloc[((dtecDat.index.values-(doyN-1))*24>(eventHr-.5))*((dtecDat.index.values-(doyN-1))*24<(eventHr+1.5)),:]
    m.scatter(site_day.pplon.values,site_day.pplat.values,c=site_day.dtec.values,latlon=True,edgecolor='none',
              s=ms,vmin=-clim,vmax=clim);
    parallels = np.arange(-80,80,5.);meridians = np.arange(0.,360.,5.);
    m.drawparallels(parallels,labels=[False,True,True,False])
    m.drawmeridians(meridians,labels=[True,False,False,True])


def plotTracks(prns,asite,raw_tec,eventHr,doyN,lat_0,lon_0,m,fsize=(10,10),ms=[15,5,8]):
    #prns = array of prn's, e.g. [20,21,29]
    plt.figure(figsize=fsize);m.drawcoastlines();m.plot(lon_0,lat_0,'r*',latlon=True,markersize=ms[0]);
    m.plot(raw_tec[0].rec_lon.values,raw_tec[0].rec_lat.values,'w^',latlon=True,markersize=ms[1]);
    m.plot(D_Kaikoura2016[1][0].loc[asite].rec_lon,D_Kaikoura2016[1][0].loc[asite].rec_lat,'b^',latlon=True,markersize=ms[2]);
    for aprn in prns:
        site_day = raw_tec[1].iloc[(raw_tec[1].prn.values==aprn),:]
        dtecDat = site_day.assign(dtec=0)
        prdat = site_day.iloc[site_day.site.values==asite]
        dtec = prdat.vtec - utec.get_lp_tec((prdat.index.values-(doyN-1))*24,prdat.vtec.values)
        dtecDat.loc[dtecDat.site.values==asite,'dtec'] = dtec
        site_day = dtecDat.iloc[(dtecDat.prn.values==aprn)*(dtecDat.site.values==asite),:]
        site_day = site_day.iloc[((site_day.index.values-(doyN-1))*24>(eventHr-.5))*((site_day.index.values-(doyN-1))*24<(eventHr+1.5)),:]
        m.scatter(site_day.pplon.values,site_day.pplat.values,c=site_day.dtec.values,latlon=True,edgecolor='none',s=5,vmin=-.1,vmax=.1);
        t_xy = m(site_day.pplon.values[0],site_day.pplat.values[0])
        plt.text(t_xy[0],t_xy[1],'PRN '+str(aprn));
    parallels = np.arange(-80,80,5.);meridians = np.arange(0.,360.,5.);
    m.drawparallels(parallels,labels=[False,True,True,False])
    m.drawmeridians(meridians,labels=[True,False,False,True])


def genHodochron(raw_data,aprn,doyN,lat_0,lon_0):
    dists_m1 = dict(); times_m1 = dict(); vtecs_m1 = dict(); dtecs_m1 = dict()
    for asite in list(raw_data[0].index):
        gtemp = raw_data[1].iloc[(raw_data[1].prn.values==aprn)*(raw_data[1].site.values==asite),:]
        gtdis = []
        for ii in np.arange(len(gtemp)):
            gtdis.append(geocalc(gtemp.iloc[ii].pplat,gtemp.iloc[ii].pplon,lat_0,lon_0))
        dsigns = (gtemp.pplat.values>lat_0)*2-1
        dists_m1[asite] = gtdis*dsigns
        times_m1[asite] = (gtemp.index.values-(doyN-1))*24
        vtecs_m1[asite] = gtemp.vtec.values
        dtecs_m1[asite] = gtemp.vtec.values - get_lp_tec((gtemp.index.values-(doyN-1))*24,gtemp.vtec.values)
    return dists_m1,times_m1,vtecs_m1,dtecs_m1


def plotHodochron(genRes,eventTime,propTime=None,ylim=[-1500,1500],clim=.1,figsize=(12,5),ms=5,nDir=True,fntsize=10):
    dists_m1,times_m1,vtecs_m1,dtecs_m1 = genRes

    plt.figure(figsize=figsize);
    for asite in list(times_m1.keys()):
        if nDir:
            plt.scatter(times_m1[asite],np.abs(dists_m1[asite]),c=dtecs_m1[asite],edgecolor='none',s=ms);
        else:
            plt.scatter(times_m1[asite],dists_m1[asite],c=dtecs_m1[asite],edgecolor='none',s=ms);
        plt.clim(-clim,clim);
    plt.plot([eventTime,eventTime],[ylim[0]-10,ylim[1]+10],'r-');
    if propTime:
        plt.plot([eventTime+propTime/60,eventTime+propTime/60],[ylim[0]-10,ylim[1]+10],'m--');
    plt.xlim(np.round(eventTime,decimals=1)-.2,np.round(eventTime,decimals=1)+1);
    plt.ylim(ylim[0],ylim[1]);cbar=plt.colorbar();
    plt.xlabel('Time (Hr)',fontsize=fntsize); plt.ylabel('Distance (km)',fontsize=fntsize);
    cbar.ax.set_title('dTEC',fontsize=fntsize);



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdaccess.framework.data_class import DataFetcherBase
from skdaccess.framework.data_class import TableWrapper
import pandas as pd
import numpy as np

class DataGenerator(DataFetcherBase):

    ''' In Development: Class for generating random data '''
    
    def __init__(self, length, *args, seed = None, final_function = None):
        '''
        Initialize Random data generator


        @param length: Number of rows to generate
        @param *args: Dictionaries containing entries: 'name',,'start', 'end', and optionally 'func'
        @param seed: Seed to use when generating random data
        @param final_function: Final function to call on random data
        '''

        self.length = length
        self.seed = seed
        self.args = args
        self.final_function = final_function

    def output(self):
        if self.seed is not None:
            np.random.seed(self.seed)
            

        new_data = dict()
        name_list = []
        for arg in self.args:
            new_data[arg['name']] = np.random.rand(self.length) * (arg['end'] - arg['start']) + arg['start']

            name_list.append(arg['name'])

            if 'func' in arg:
                new_data[arg['name']] = arg['func'](new_data[arg['name']])


        if self.final_function is not None:
            new_data = pd.DataFrame.from_dict(new_data)
            new_data, updated_column_names = self.final_function(new_data)
            if updated_column_names is not None:
                default_columns = updated_column_names


        data = {'generated_data' : new_data}
        data_wrapper = TableWrapper(data, default_columns = name_list)
        
        return data_wrapper



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem
import numpy as np
import matplotlib.pyplot as plt
import math

class Plotter(PipelineItem):
    '''
    Make a plot of table data
    '''
    def __init__(self, str_description, column_names=None, error_column_names = None, num_columns = 3, width=13, height=4, columns_together=False,
                 annotate_column = None, annotate_data = None, xlim = None, ylim = None, **kwargs):
        '''
        Initialize Plotter
        
        @param str_description: String describing accumulator
        @param column_names: Columns to be plot
        @param error_column_names Columns containing uncertainties to be plot, no errorbars if None
        @param num_columns: Number of columns to use when plotting data
        @param width: Total width of all columns combined
        @param height: Height of single row of plots
        @param columns_together: If true, plot the columns on the same graph
        @param annotate_column: Column of annotation data to use for annotation
        @param annotate_data: Annotation data
        @param xlim: The x limit
        @param ylim: The y limit
        @param **kwargs: Any additional keyword arguments are passed on to matplotlib
        '''
        self.xlim = xlim
        self.ylim = ylim
        self.kwargs = kwargs
        self.num_columns = num_columns
        self.height = height
        self.width = width
        self.column_names = column_names
        self.annotate_column = annotate_column
        self.annotate_data = annotate_data
        
        self.error_column_names = error_column_names
        self.columns_together = columns_together
        super(Plotter, self).__init__(str_description, [])
    
    def process(self, obj_data):
        '''
        Plot each column in obj_data

        @param obj_data: Data Wrapper
        '''

        if self.column_names == None:
            column_names = obj_data.getDefaultColumns()
        else:
            column_names = self.column_names

        width = self.width
        height = self.height

        # Determine total number of figures needed
        if self.columns_together == False:
            num_figures = obj_data.getLength() * len(column_names)
        else:
            num_figures =  obj_data.getLength()
        if num_figures > 0:

            # Determine number of rows and height needed to plot all figures
            rows = math.ceil( num_figures / self.num_columns)
            height *= rows

            figure = plt.figure()
            figure.set_size_inches(width, height, True)

            if self.xlim != None:
                plt.xlim(*self.xlim)

            if self.ylim != None:
                plt.ylim(*self.ylim)
            
            num = 0
            # Main loop that iterates over all data
            for label, data in obj_data.getIterator():
                if self.columns_together == True:
                    num += 1
                    
                # Plotting with errorbars
                if self.error_column_names != None:
                    for column, err_column in zip(column_names, self.error_column_names):
                        if self.columns_together == False:
                            num += 1

                        plt.subplot(rows, self.num_columns, num)
                        plt.title(label)
                        plt.ylabel(column)

                        plt.xticks(rotation=45)
                        plt.errorbar(np.array(data.index),np.array(data[column]), yerr=np.array(data[err_column]), **self.kwargs)

                        if self.annotate_column is not None:
                            try:
                                for vline in self.annotate_data[label][self.annotate_column]:
                                    plt.axvline(vline,color='black',linewidth=3,alpha=0.5)
                            except KeyError:
                                pass
                                # print('cannot find info')
                        
                        elif self.annotate_data is not None:
                            try:
                                for vline in self.annotate_data[label]:
                                    plt.axvline(vline,color='black',linewidth=3,alpha=0.5)
                            except KeyError:
                                pass
                                # print('cannot find info')

                # Plotting without errorbars
                else:
                    for column in column_names:
                        if self.columns_together == False:
                            num += 1
                        
                        plt.subplot(rows, self.num_columns, num)
                        plt.title(label)
                        plt.ylabel(column)

                        plt.xticks(rotation=45)

                        plt.plot(data[column], **self.kwargs)
                        if self.annotate_column is not None:
                            try:
                                for vline in self.annotate_data[label][self.annotate_column]:
                                    plt.axvline(vline,color='black',linewidth=3,alpha=0.5)
                            except KeyError:
                                pass

                        elif self.annotate_data is not None:
                            try:
                                for vline in self.annotate_data[label]:
                                    plt.axvline(vline,color='black',linewidth=3,alpha=0.5)
                            except KeyError:
                                pass
                                
                        

            # Tight layout usually dispalys nicer
            plt.tight_layout()

            # If run_id is > -1, display run number on figure
            if(obj_data.run_id > -1):
                figure.suptitle( "Run: " + str(obj_data.run_id), y=1.02)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import numpy as np
import pandas as pd
from astropy.cosmology import FlatLambdaCDM
from scipy.integrate import quad
from scipy.optimize import root
from skdaccess.framework.data_class import DataFetcherBase, TableWrapper

from skdiscovery.utilities.patterns.astro_tools import nfw, move_point


class CatalogGenerator(DataFetcherBase):
    '''
    *In Development* Generates galaxy catalogs for use in DiscoveryPipeline
    '''

    def __init__(self, ap_paramList, ra1,dec1, ra2, dec2, background_density, z):
        '''
        @param ap_paramList[seed]: Seed for random number generator
        @param ra1:  Left right ascension
        @param dec1: Bottom declination
        @param ra2:  Right right ascension
        @param dec2: Top declination
        @param background_density: galaxy background density in galaxies/square degree
        @param z: Redshift of galaxy cluster
        '''

        self.ra1 = ra1
        self.dec1 = dec1
        self.ra2 = ra2
        self.dec2 = dec2
        self.background_density = background_density
        self.z = z

        self.__H0 = 70
        self.__Omega_m = 0.3

        self.__R0 = 2
        self.__Rs = 0.15 / 0.7
        self.__Rcore = 0.1 / 0.7
        self.__norm = 1/quad(nfw,0,self.__R0,args=(1,self.__Rs,self.__Rcore))[0]

        super(CatalogGenerator, self).__init__(ap_paramList)
    

    def output(self):
        ''' 
        Generates galaxy catalog

        @return DataWrapper: Table data wrapper of galaxy catalog
        '''

        if len(self.ap_paramList) > 0:
            seed = self.ap_paramList[0]()
            np.random.seed(seed)

        else:
            seed = None

        # Generating a background for the whole sky, and then selecting galaxies
        # from that background
        num_background_galaxies = round(self.background_density * (4 * np.pi) * (180./np.pi)**2)

        # Generate background galaxies over an entire sphere
        full_ra = np.random.rand(num_background_galaxies) * 360.
        full_dec = np.arccos(2*np.random.rand(num_background_galaxies)-1) * 180. / np.pi - 90.

        # Make data frame and select galaxies in our box
        galaxies = pd.DataFrame.from_dict({'RA' : full_ra, 'Dec' : full_dec})

        galaxies = galaxies[ np.logical_and.reduce( (galaxies['RA'] > self.ra1, galaxies['RA'] < self.ra2,
                                                     galaxies['Dec'] > self.dec1, galaxies['Dec'] < self.dec2) ) ]

        galaxies['Cluster_ID'] = -1        


        # Now generating a cluster
        num_galaxies = 100

        cluster_ra = np.mean([self.ra1,self.ra2])
        cluster_dec = np.mean([self.dec1, self.dec2])

        cosmo = FlatLambdaCDM(self.__H0, self.__Omega_m)

        distance = cosmo.comoving_distance(self.z).to('Mpc').value

        radial_positions = self.inverse_nfw_cumulative(np.random.rand(num_galaxies))
        bearings = np.random.rand(num_galaxies) * 360.

        # Converting radial comoving distances to angular seperations
        angular_distances = radial_positions / distance * 180./np.pi

        # Generate the RA and Dec for each galaxy by moving it in random direction
        # the radial distance away from the center
        
        cluster_ra, cluster_dec = move_point(cluster_ra, cluster_dec, angular_distances, bearings)
        cluster_galaxies = pd.DataFrame.from_dict( {'RA' : cluster_ra, 'Dec' : cluster_dec} )

        cluster_galaxies['Cluster_ID'] = int(0)

        galaxies = pd.concat([galaxies, cluster_galaxies], axis=0).set_index(np.arange(len(galaxies) + len(cluster_galaxies)))

        data_wrapper = TableWrapper({'Cluster_Catalog_01' : galaxies}, default_columns=['RA','Dec'])
        
        return data_wrapper




    def nfw_cumulative(self,R):
        ''' 
        Cumulative radial NFW distribution

        @param R: Radius
        @return float: Probability of being within R
        '''
        
        R0 = self.__R0
        norm = self.__norm
        Rs = self.__Rs
        Rcore = self.__Rcore

        def integrate(z):
            return quad(nfw,0,z,args=(norm,Rs,Rcore))[0]

        if np.isscalar(R):
            R = np.array([float(R)])
        else:
            R = R.astype(np.float)

        res = np.piecewise(R, [R < 0.0, np.logical_and(R >= 0.0, R < R0), R >= R0], 
                           [lambda z: z, 
                            lambda z: np.fromiter(map(integrate,z),np.float),
                            lambda z: z/R0])
        if len(res) == 1:
            return res[0]
        else:
            return res

    def inverse_nfw_cumulative(self, p):
        '''
        inverse of radial nfw cumulative distribution

        @param p: Probability
        @return float: Radius corresponding to probability p 
        '''
        R0 = self.__R0

        def get_root(p):
            return root(lambda z: self.nfw_cumulative(z) - p,R0/2).x[0]
        if np.isscalar(p):
            get_root(p)
        else:
            return np.fromiter(map(get_root,p),np.float, count=len(p))



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import matplotlib.pyplot as plt
import numpy as np
import skdaccess.utilities.pbo_util as pbo_utils
import skdiscovery.data_structure.series.analysis.mogi as mogi
from mpl_toolkits.basemap import Basemap
from skdiscovery.data_structure.framework import PipelineItem

from skdiscovery.utilities.patterns import pbo_tools


class GPSHPlotter(PipelineItem):
    ''' 
    Plots results from General_Component_Analysis, for the GPS horizontal or vertical components
    '''
    
    def __init__(self, str_description, comp_name, mogi_name = None, pca_dir = 'H', pca_comp = 0, scaleFactor = 2.5, offset=.15, KF_tau = 0, errorEllipses = False, map_resolution='i'):
        '''
        Initialize GPHSHPlotter

        @param str_description: String describing accumulator
        @param comp_name: Name of the GPCA results for accessing the GPCA output
        @param mogi_name: Name of the Mogi results (optional)
        @param pca_dir: PCA direction to plot, horizontal (H) or vertical (V)
        @param pca_comp: The PCA component that will be plotted
        @param scaleFactor: Scale factor for arrows
        @param offset: Offset for plotting larger area on map
        @param KF_tau: Tau used in kalman filter
        @param errorEllipses: Boolean indicating whether or not to plot errorEllipses
        '''

        self.dir_sign = 0        
        self.pca_dir = pca_dir
        self.pca_comp = pca_comp
        self.scaleFactor = scaleFactor
        self.offset = offset
        self.errorE = errorEllipses
        self.KF_tau = KF_tau
        self._bmap_res = map_resolution
        self.comp_name = comp_name
        self.mogi_name = mogi_name
        super(GPSHPlotter, self).__init__(str_description)
        
    def process(self, obj_data):
        ''' 
        Plot the General Component Analysis results present stored in obj_data. 

        Saves the basemap in obj_data results.
        
        @param obj_data: Data Wrapper that holds component analysis HPCA
        '''
        HPCA_name = self.comp_name
        Mogi_name = self.mogi_name
        pca_comp  = self.pca_comp

        plt.figure()

        meta_data = obj_data.info()
        try:
            station_list = obj_data.get().minor_axis
        except AttributeError:
            station_list = list(obj_data.get().keys())

        lat_range, lon_range = pbo_utils.getLatLonRange(meta_data, station_list)
        coord_list = pbo_utils.getStationCoords(meta_data, station_list)

        # Create a map projection of area
        offset = self.offset
        bmap = Basemap(llcrnrlat=lat_range[0] - offset, urcrnrlat=lat_range[1] + offset, llcrnrlon=lon_range[0] - offset, urcrnrlon=lon_range[1] + offset,
                       projection='gnom', lon_0=np.mean(lon_range), lat_0=np.mean(lat_range), resolution=self._bmap_res)

        # bmap.fillcontinents(color='white')
        bmap.drawmapboundary(fill_color='white')

        # Draw just coastlines, no lakes
        for i,cp in enumerate(bmap.coastpolygons):
             if bmap.coastpolygontypes[i]<2:
                bmap.plot(cp[0],cp[1],'k-')

        parallels = np.arange(np.round(lat_range[0]-offset,decimals=1),np.round(lat_range[1]+offset,decimals=1),.1)
        meridians = np.arange(np.round(lon_range[0]-offset,decimals=1),np.round(lon_range[1]+offset,decimals=1),.1)

        bmap.drawmeridians(meridians, labels=[0,0,0,1],fontsize=14)
        bmap.drawparallels(parallels, labels=[1,0,0,0],fontsize=14)

        
        pca_results = obj_data.getResults()[HPCA_name]
        pca = pca_results['CA']

        lonscale = 1
        latscale = 1
        scaleFactor = self.scaleFactor

        if self.pca_dir == 'V':
            station_lat_list, station_lon_list, ev_lat_list, ev_lon_list, dir_sign = pbo_tools.dirEigenvectors(coord_list, pca.components_[pca_comp],pdir='V')
            self.dir_sign = dir_sign            
            pca_results['Projection'] *= dir_sign
            ev_lat_list *= latscale
            ev_lon_list *= lonscale
                
            # Plot station coords
            for coord in coord_list:
                bmap.plot(coord[1], coord[0], 'bo', markersize=8, latlon=True)
                x,y = bmap(coord[1], coord[0])
                plt.text(x-(1+np.sign(ev_lon_list[coord_list.index(coord)]))*900+250,
                         y-(1+np.sign(ev_lat_list[coord_list.index(coord)]))*100+450,
                         station_list[coord_list.index(coord)],fontsize=14)
    
            bmap.quiver(station_lon_list, station_lat_list, ev_lon_list, ev_lat_list, latlon=True, scale = scaleFactor)
            
            ax_x = plt.gca().get_xlim()
            ax_y = plt.gca().get_ylim()
            x,y = bmap(ax_x[0]+.1*(ax_x[1]-ax_x[0]), ax_y[0]+.1*(ax_y[1]-ax_y[0]),inverse=True)
            bmap.quiver(x, y, 0, .2, latlon=True, scale = scaleFactor, headwidth=3,headlength=3)
            plt.text(ax_x[0]+.1*(ax_x[1]-ax_x[0])-650, ax_y[0]+.1*(ax_y[1]-ax_y[0])-1000,'20%', fontsize=14)
            
        else:
            station_lat_list, station_lon_list, ev_lat_list, ev_lon_list, dir_sign = pbo_tools.dirEigenvectors(coord_list, pca.components_[pca_comp])
            self.dir_sign = dir_sign            
            pca_results['Projection'] *= dir_sign
            ev_lat_list *= latscale
            ev_lon_list *= lonscale
                
            # Plot station coords
            for coord in coord_list:
                bmap.plot(coord[1], coord[0], 'bo', markersize=8, latlon=True)
                x,y = bmap(coord[1], coord[0])
                plt.text(x-(1+np.sign(ev_lon_list[coord_list.index(coord)]))*900+250,
                         y-(1+np.sign(ev_lat_list[coord_list.index(coord)]))*800+450,
                         station_list[coord_list.index(coord)], fontsize=14)
    
            bmap.quiver(station_lon_list, station_lat_list, ev_lon_list, ev_lat_list, latlon=True, scale = scaleFactor)
            
            ax_x = plt.gca().get_xlim()
            ax_y = plt.gca().get_ylim()
            x,y = bmap(ax_x[0]+.1*(ax_x[1]-ax_x[0]), ax_y[0]+.1*(ax_y[1]-ax_y[0]),inverse=True)
            bmap.quiver(x, y, 0, .2, latlon=True, scale = scaleFactor, headwidth=3,headlength=3)
            plt.text(ax_x[0]+.1*(ax_x[1]-ax_x[0])-650, ax_y[0]+.1*(ax_y[1]-ax_y[0])-1000,'20%', fontsize=14)
    
            # Plotting Mogi source
            if Mogi_name != None:
                mogi_res = obj_data.getResults()[Mogi_name]
                bmap.plot(mogi_res['lon'], mogi_res['lat'], "g^", markersize = 10, latlon=True)
                mogi_x_disp, mogi_y_disp = mogi.MogiVectors(mogi_res,station_lat_list,station_lon_list)
                bmap.quiver(station_lon_list, station_lat_list, mogi_x_disp*dir_sign, mogi_y_disp*dir_sign,
                            latlon=True, scale=scaleFactor,color='red')
                            
            # Plot error ellipses for the PCA
            if self.errorE:
                ax = plt.gca()
                yScale = (bmap.urcrnrlat - bmap.llcrnrlat)/scaleFactor
                xScale = (bmap.urcrnrlon - bmap.llcrnrlon)/scaleFactor
                midY = (bmap.urcrnrlat + bmap.llcrnrlat)/2
                midX = (bmap.urcrnrlon + bmap.llcrnrlon)/2
                from matplotlib.patches import Ellipse
 
                n=len(pca_results['Projection'][:,0])
                tau=self.KF_tau
                delta_t = np.arange(-(n-1),n+1)
                mseq = (1-np.abs(delta_t)/n)
                rdelt = np.exp(-np.abs(delta_t)/tau)
                neff = n/np.sum(mseq*rdelt)
                eigval = pca.explained_variance_
                aaTs = [np.outer(pca.components_[ii,:],pca.components_[ii,:].T) for ii in range(pca.components_.shape[0])]
                VVs = [eigval[ii]/neff*np.sum([eigval[k]/(eigval[k]-eigval[ii])**2*aaTs[k] for k in (j for j in range(pca.components_.shape[0]) if j != ii)],axis=0) for ii in range(pca.components_.shape[0])]
                sigmas = np.diag(VVs[0])**(1/2)                
                
                for kk in range(len(station_lon_list)):
                    vlon = ev_lon_list[kk]
                    vlat = ev_lat_list[kk]
                    slon = station_lon_list[kk]
                    slat = station_lat_list[kk]
                    Elat = sigmas[2*kk]
                    Elon = sigmas[2*kk+1]
                    cir_w, cir_h = np.array(bmap(midX+Elon/scaleFactor,midY+Elat/scaleFactor*.85))-np.array(bmap(midX,midY))
                
                    x,y = bmap(slon+vlon*xScale*.95,slat+vlat*yScale*.85)
                    # if need to rotate ellipse, np.arctan2(vlat,vlon)*180/np.pi
                    etest = Ellipse(xy=(x,y),width=cir_w,height=cir_h,angle=0,
                                    edgecolor='k',fc='w',lw=1,zorder=-1)
                    ax.add_artist(etest);
                    
                           
        obj_data.addResult(self.str_description, bmap)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from skdiscovery.data_structure.framework import PipelineItem
from skdaccess.framework.param_class import AutoParam
import numpy as np

class SnowRemover(PipelineItem):
    ''' Removes data with snow errors '''

    def __init__(self,str_description,ap_paramList = [AutoParam(1.5)],column_name='dN', snow_column='Snow'):
        '''
        Initialize snow remover for use on table data.

        @param str_description: String describing filter
        @param ap_paramList[sigma_clip]: remove station if the stddev of snowdays is sigma_clip 
                                         times greater than non-snow days, default 1.5
        @param column_name: Name of column to check
        @param snow_column: Name of snow column to determine snowdays/non snow days
        '''
        self.column_name = column_name
        self.snow_column = snow_column


        super(SnowRemover, self).__init__(str_description, ap_paramList)


    def process(self, obj_data):
        ''' 
        Removes table data with large snow errors

        @param obj_data: Input DataWrapper, will be modified in place
        '''

        bad_stations = []
        sigma_multiplier = self.ap_paramList[0]()

        for label, data in obj_data.getIterator():

            if len(data[data[self.snow_column]==4]) > 0 and len(data[data[self.snow_column]==2]) > 0:
                snow = data[data[self.snow_column]==4].loc[:,self.column_name]
                no_snow = data[data[self.snow_column]==2].loc[:,self.column_name]

                non_snow_std = np.nanstd(no_snow)
                snow_std = np.nanstd(snow)

                if snow_std > sigma_multiplier * non_snow_std:
                    bad_stations.append(label)

        if len(bad_stations) > 0:
            obj_data.removeFrames(bad_stations)

